@@ -15,16 +15,21 @@
  */
 
 #include <assert.h>
+#include <stdint.h>
 #include <uv.h>
 #include <contrib/cleanup.h>
 #include <libknot/descriptor.h>
 
 #include "lib/cache.h"
 #include "lib/cdb.h"
+#include "lib/utils.h"
 #include "daemon/bindings.h"
 #include "daemon/worker.h"
 #include "daemon/tls.h"
 
+#define xstr(s) str(s)
+#define str(s) #s
+
 /** @internal Annotate for static checkers. */
 KR_NORETURN int lua_error (lua_State *L);
 
@@ -184,7 +189,7 @@ static int net_listen_addrs(lua_State *L, int port, int flags)
 		struct engine *engine = engine_luaget(L);
 		int ret = network_listen(&engine->net, str, port, flags);
 		if (ret != 0) {
-			kr_log_info("[system] bind to '%s#%d' %s\n",
+			kr_log_info("[system] bind to '%s@%d' %s\n",
 					str, port, kr_strerror(ret));
 		}
 		return ret == 0;
@@ -223,7 +228,7 @@ static int net_listen(lua_State *L)
 	int n = lua_gettop(L);
 	if (n < 1 || n > 3) {
 		format_error(L, "expected one to three arguments; usage:\n"
-			"net.listen(addressses, [port = 53, flags = {tls = (port == 853)}])\n");
+			     "net.listen(addressses, [port = " xstr(KR_DNS_PORT) ", flags = {tls = (port == " xstr(KR_DNS_TLS_PORT) ")}])\n");
 		lua_error(L);
 	}
 
@@ -300,10 +305,9 @@ static int net_interfaces(lua_State *L)
 		char *p = buf;
 		memset(buf, 0, sizeof(buf));
 		for (unsigned k = 0; k < sizeof(iface.phys_addr); ++k) {
-			sprintf(p, "%.2x:", iface.phys_addr[k] & 0xff);
+			sprintf(p, "%s%.2x", k > 0 ? ":" : "", iface.phys_addr[k] & 0xff);
 			p += 3;
 		}
-		*(p - 1) = '\0';
 		lua_pushstring(L, buf);
 		lua_setfield(L, -2, "mac");
 
@@ -326,7 +330,7 @@ static int net_bufsize(lua_State *L)
 	}
 	int bufsize = lua_tointeger(L, 1);
 	if (bufsize < 512 || bufsize > UINT16_MAX) {
-		format_error(L, "bufsize must be within <512, 65535>");
+		format_error(L, "bufsize must be within <512, " xstr(UINT16_MAX) ">");
 		lua_error(L);
 	}
 	knot_edns_set_payload(opt_rr, (uint16_t) bufsize);
@@ -346,7 +350,7 @@ static int net_pipeline(lua_State *L)
 	}
 	int len = lua_tointeger(L, 1);
 	if (len < 0 || len > UINT16_MAX) {
-		format_error(L, "tcp_pipeline must be within <0, 65535>");
+		format_error(L, "tcp_pipeline must be within <0, " xstr(UINT16_MAX) ">");
 		lua_error(L);
 	}
 	worker->tcp_pipeline_max = len;
@@ -400,27 +404,100 @@ static int net_tls_padding(lua_State *L)
 
 	/* Only return current padding. */
 	if (lua_gettop(L) == 0) {
-		if (engine->resolver.tls_padding == 0) {
-			return -1;
+		if (engine->resolver.tls_padding < 0) {
+			lua_pushboolean(L, true);
+			return 1;
+		} else if (engine->resolver.tls_padding == 0) {
+			lua_pushboolean(L, false);
+			return 1;
 		}
 		lua_pushinteger(L, engine->resolver.tls_padding);
 		return 1;
 	}
 
-	if ((lua_gettop(L) != 1) || !lua_isnumber(L, 1)) {
-		lua_pushstring(L, "net.tls_padding takes one numeric parameter: (\"padding\")");
+	if ((lua_gettop(L) != 1)) {
+		lua_pushstring(L, "net.tls_padding takes one parameter: (\"padding\")");
 		lua_error(L);
 	}
-	int padding = lua_tointeger(L, 1);
-	if ((padding < 0) || (padding > MAX_TLS_PADDING)) {
-		lua_pushstring(L, "net.tls_padding parameter has to be a number between <0, 4096>");
+	if (lua_isboolean(L, 1)) {
+		bool x = lua_toboolean(L, 1);
+		if (x) {
+			engine->resolver.tls_padding = -1;
+		} else {
+			engine->resolver.tls_padding = 0;
+		}
+	} else if (lua_isnumber(L, 1)) {
+		int padding = lua_tointeger(L, 1);
+		if ((padding < 0) || (padding > MAX_TLS_PADDING)) {
+			lua_pushstring(L, "net.tls_padding parameter has to be true, false, or a number between <0, " xstr(MAX_TLS_PADDING) ">");
+			lua_error(L);
+		}
+		engine->resolver.tls_padding = padding;
+	} else {
+		lua_pushstring(L, "net.tls_padding parameter has to be true, false, or a number between <0, " xstr(MAX_TLS_PADDING) ">");
 		lua_error(L);
 	}
-	engine->resolver.tls_padding = padding;
 	lua_pushboolean(L, true);
 	return 1;
 }
 
+static int net_outgoing(lua_State *L, int family)
+{
+	struct worker_ctx *worker = wrk_luaget(L);
+	union inaddr *addr;
+	if (family == AF_INET)
+		addr = (union inaddr*)&worker->out_addr4;
+	else
+		addr = (union inaddr*)&worker->out_addr6;
+
+	if (lua_gettop(L) == 0) { /* Return the current value. */
+		if (addr->ip.sa_family == AF_UNSPEC) {
+			lua_pushnil(L);
+			return 1;
+		}
+		if (addr->ip.sa_family != family) {
+			assert(false);
+			lua_error(L);
+		}
+		char addr_buf[INET6_ADDRSTRLEN];
+		int err;
+		if (family == AF_INET)
+			err = uv_ip4_name(&addr->ip4, addr_buf, sizeof(addr_buf));
+		else
+			err = uv_ip6_name(&addr->ip6, addr_buf, sizeof(addr_buf));
+		if (err)
+			lua_error(L);
+		lua_pushstring(L, addr_buf);
+		return 1;
+	}
+
+	if ((lua_gettop(L) != 1) || (!lua_isstring(L, 1) && !lua_isnil(L, 1))) {
+		format_error(L, "net.outgoing_vX takes one address string parameter or nil");
+		lua_error(L);
+	}
+
+	if (lua_isnil(L, 1)) {
+		addr->ip.sa_family = AF_UNSPEC;
+		return 1;
+	}
+
+	const char *addr_str = lua_tostring(L, 1);
+	int err;
+	if (family == AF_INET)
+		err = uv_ip4_addr(addr_str, 0, &addr->ip4);
+	else
+		err = uv_ip6_addr(addr_str, 0, &addr->ip6);
+	if (err) {
+		format_error(L, "net.outgoing_vX: failed to parse the address");
+		lua_error(L);
+	}
+	lua_pushboolean(L, true);
+	return 1;
+}
+
+static int net_outgoing_v4(lua_State *L) { return net_outgoing(L, AF_INET); }
+static int net_outgoing_v6(lua_State *L) { return net_outgoing(L, AF_INET6); }
+
 int lib_net(lua_State *L)
 {
 	static const luaL_Reg lib[] = {
@@ -432,6 +509,8 @@ int lib_net(lua_State *L)
 		{ "tcp_pipeline", net_pipeline },
 		{ "tls",          net_tls },
 		{ "tls_padding",  net_tls_padding },
+		{ "outgoing_v4",  net_outgoing_v4 },
+		{ "outgoing_v6",  net_outgoing_v6 },
 		{ NULL, NULL }
 	};
 	register_lib(L, "net", lib);
@@ -518,7 +597,7 @@ static int cache_max_ttl(lua_State *L)
 		uint32_t min = cache->ttl_min;
 		int64_t ttl = lua_tonumber(L, 1);
 		if (ttl < 0 || ttl <= min || ttl > UINT32_MAX) {
-			format_error(L, "max_ttl must be larger than minimum TTL, and in range <1, UINT32_MAX>'");
+			format_error(L, "max_ttl must be larger than minimum TTL, and in range <1, " xstr(UINT32_MAX) ">'");
 			lua_error(L);
 		}
 		cache->ttl_max = ttl;
@@ -542,7 +621,7 @@ static int cache_min_ttl(lua_State *L)
 		uint32_t max = cache->ttl_max;
 		int64_t ttl = lua_tonumber(L, 1);
 		if (ttl < 0 || ttl >= max || ttl > UINT32_MAX) {
-			format_error(L, "min_ttl must be smaller than maximum TTL, and in range <0, UINT32_MAX>'");
+			format_error(L, "min_ttl must be smaller than maximum TTL, and in range <0, " xstr(UINT32_MAX) ">'");
 			lua_error(L);
 		}
 		cache->ttl_min = ttl;
@@ -563,7 +642,14 @@ static int cache_open(lua_State *L)
 
 	/* Select cache storage backend */
 	struct engine *engine = engine_luaget(L);
-	unsigned cache_size = lua_tonumber(L, 1);
+
+	lua_Number csize_lua = lua_tonumber(L, 1);
+	if (!(csize_lua >= 8192 && csize_lua < SIZE_MAX)) { /* min. is basically arbitrary */
+		format_error(L, "invalid cache size specified, it must be in range <8192, " xstr(SIZE_MAX)  ">");
+		lua_error(L);
+	}
+	size_t cache_size = csize_lua;
+
 	const char *conf = n > 1 ? lua_tostring(L, 2) : NULL;
 	const char *uri = conf;
 	const struct kr_cdb_api *api = cache_select(engine, &conf);
@@ -1102,7 +1188,10 @@ static int wrk_resolve(lua_State *L)
 		lua_error(L);
 	}
 	uint8_t dname[KNOT_DNAME_MAXLEN];
-	knot_dname_from_str(dname, lua_tostring(L, 1), sizeof(dname));
+	if (!knot_dname_from_str(dname, lua_tostring(L, 1), sizeof(dname))) {
+		lua_pushstring(L, "invalid qname");
+		lua_error(L);
+	};
 	/* Check class and type */
 	uint16_t rrtype = lua_tointeger(L, 2);
 	if (!lua_isnumber(L, 2)) {
