@@ -21,6 +21,7 @@
 #include <unistd.h>
 #include <grp.h>
 #include <pwd.h>
+#include <sys/param.h>
 #include <zscanner/scanner.h>
 
 #include "daemon/engine.h"
@@ -43,6 +44,13 @@ KR_NORETURN int lua_error (lua_State *L);
 /* Cleanup engine state every 5 minutes */
 const size_t CLEANUP_TIMER = 5*60*1000;
 
+/* Execute byte code */
+#define l_dobytecode(L, arr, len, name) \
+	(luaL_loadbuffer((L), (arr), (len), (name)) || lua_pcall((L), 0, LUA_MULTRET, 0))
+/** Load file in a sandbox environment. */
+#define l_dosandboxfile(L, filename) \
+	(luaL_loadfile((L), (filename)) || engine_pcall((L), 0))
+
 /*
  * Global bindings.
  */
@@ -201,6 +209,63 @@ static int l_hostname(lua_State *L)
 	return 1;
 }
 
+char *engine_get_moduledir(struct engine *engine) {
+	return engine->moduledir;
+}
+
+int engine_set_moduledir(struct engine *engine, const char *moduledir) {
+	if (!engine || !moduledir) {
+		return kr_error(EINVAL);
+	}
+
+	char *new_moduledir = strdup(moduledir);
+	if (!new_moduledir) {
+		return kr_error(ENOMEM);
+	}
+	if (engine->moduledir) {
+		free(engine->moduledir);
+	}
+	engine->moduledir = new_moduledir;
+
+	/* Use module path for including Lua scripts */
+	char l_paths[MAXPATHLEN] = { 0 };
+	/* Save original package.path to package._path */
+	snprintf(l_paths, MAXPATHLEN - 1,
+		 "if package._path == nil then package._path = package.path end\n"
+		 "package.path = '%s/?.lua;%s/?/init.lua;'..package._path\n",
+		 new_moduledir,
+		 new_moduledir);
+
+	int ret = l_dobytecode(engine->L, l_paths, strlen(l_paths), "");
+	if (ret != 0) {
+		lua_pop(engine->L, 1);
+		return ret;
+	}
+	return 0;
+}
+
+/** Return hostname. */
+static int l_moduledir(lua_State *L)
+{
+	struct engine *engine = engine_luaget(L);
+	if (lua_gettop(L) == 0) {
+		lua_pushstring(L, engine_get_moduledir(engine));
+		return 1;
+	}
+	if ((lua_gettop(L) != 1) || !lua_isstring(L, 1)) {
+		lua_pushstring(L, "moduledir takes at most one parameter: (\"directory\")");
+		lua_error(L);
+	}
+
+	if (engine_set_moduledir(engine, lua_tostring(L, 1)) != 0) {
+		lua_pushstring(L, "setting moduledir failed");
+		lua_error(L);
+	}
+
+	lua_pushstring(L, engine_get_moduledir(engine));
+	return 1;
+}
+
 /** Get/set context option. */
 static int l_option(lua_State *L)
 {
@@ -496,8 +561,8 @@ static int init_resolver(struct engine *engine)
 		return kr_error(ENOMEM);
 	}
 	knot_edns_init(engine->resolver.opt_rr, KR_EDNS_PAYLOAD, 0, KR_EDNS_VERSION, engine->pool);
-	/* Set default TLS padding */
-	engine->resolver.tls_padding = KR_DEFAULT_TLS_PADDING;
+	/* Use default TLS padding */
+	engine->resolver.tls_padding = -1;
 	/* Set default root hints */
 	kr_zonecut_init(&engine->resolver.root_hints, (const uint8_t *)"", engine->pool);
 	kr_zonecut_set_sbelt(&engine->resolver, &engine->resolver.root_hints);
@@ -532,6 +597,8 @@ static int init_state(struct engine *engine)
 	lua_setglobal(engine->L, "quit");
 	lua_pushcfunction(engine->L, l_hostname);
 	lua_setglobal(engine->L, "hostname");
+	lua_pushcfunction(engine->L, l_moduledir);
+	lua_setglobal(engine->L, "moduledir");
 	lua_pushcfunction(engine->L, l_verbose);
 	lua_setglobal(engine->L, "verbose");
 	lua_pushcfunction(engine->L, l_option);
@@ -548,8 +615,6 @@ static int init_state(struct engine *engine)
 	lua_setglobal(engine->L, "tojson");
 	lua_pushcfunction(engine->L, l_map);
 	lua_setglobal(engine->L, "map");
-	lua_pushliteral(engine->L, MODULEDIR);
-	lua_setglobal(engine->L, "moduledir");
 	lua_pushlightuserdata(engine->L, engine);
 	lua_setglobal(engine->L, "__engine");
 	return kr_ok();
@@ -643,6 +708,8 @@ void engine_deinit(struct engine *engine)
 	array_clear(engine->ipc_set);
 	kr_ta_clear(&engine->resolver.trust_anchors);
 	kr_ta_clear(&engine->resolver.negative_anchors);
+	free(engine->hostname);
+	free(engine->moduledir);
 }
 
 int engine_pcall(lua_State *L, int argc)
@@ -685,21 +752,9 @@ int engine_ipc(struct engine *engine, const char *expr)
 	}
 }
 
-/* Execute byte code */
-#define l_dobytecode(L, arr, len, name) \
-	(luaL_loadbuffer((L), (arr), (len), (name)) || lua_pcall((L), 0, LUA_MULTRET, 0))
-/** Load file in a sandbox environment. */
-#define l_dosandboxfile(L, filename) \
-	(luaL_loadfile((L), (filename)) || engine_pcall((L), 0))
-
 static int engine_loadconf(struct engine *engine, const char *config_path)
 {
-	/* Use module path for including Lua scripts */
-	static const char l_paths[] = "package.path = '" MODULEDIR "/?.lua;'..package.path";
-	int ret = l_dobytecode(engine->L, l_paths, sizeof(l_paths) - 1, "");
-	if (ret != 0) {
-		lua_pop(engine->L, 1);
-	}
+	int ret = 0;
 	/* Init environment */
 	static const char sandbox_bytecode[] = {
 		#include "daemon/lua/sandbox.inc"
@@ -710,10 +765,10 @@ static int engine_loadconf(struct engine *engine, const char *config_path)
 		return kr_error(ENOEXEC);
 	}
 	/* Load config file */
-	if (strcmp(config_path, "-") == 0) {
-		return ret; /* No config, no defaults. */
-	}
-	if(access(config_path, F_OK ) != -1 ) {
+	if (config_path) {
+		if (strcmp(config_path, "-") == 0) {
+			return ret; /* No config and no defaults. */
+		}
 		ret = l_dosandboxfile(engine->L, config_path);
 	}
 	if (ret == 0) {
@@ -771,15 +826,20 @@ void engine_stop(struct engine *engine)
 	uv_stop(uv_default_loop());
 }
 
-/** Register module properties in Lua environment */
+/** Register module properties in Lua environment, if any. */
 static int register_properties(struct engine *engine, struct kr_module *module)
 {
+	if (!module->config && !module->props) {
+		return kr_ok();
+	}
 	lua_newtable(engine->L);
 	if (module->config != NULL) {
 		REGISTER_MODULE_CALL(engine->L, module, module->config, "config");
 	}
-	for (struct kr_prop *p = module->props; p && p->name; ++p) {
-		if (p->cb != NULL && p->name != NULL) {
+
+	const struct kr_prop *p = module->props == NULL ? NULL : module->props();
+	for (; p && p->name; ++p) {
+		if (p->cb != NULL) {
 			REGISTER_MODULE_CALL(engine->L, module, p->cb, p->name);
 		}
 	}
@@ -831,7 +891,7 @@ int engine_register(struct engine *engine, const char *name, const char *precede
 		return kr_error(ENOMEM);
 	}
 	module->data = engine;
-	int ret = kr_module_load(module, name, NULL);
+	int ret = kr_module_load(module, name, engine->moduledir);
 	/* Load Lua module if not a binary */
 	if (ret == kr_error(ENOENT)) {
 		ret = ffimodule_register_lua(engine, module, name);
@@ -865,12 +925,7 @@ int engine_register(struct engine *engine, const char *name, const char *precede
 		}
 	}
 
-	/* Register properties */
-	if (module->props || module->config) {
-		return register_properties(engine, module);
-	}
-
-	return kr_ok();
+	return register_properties(engine, module);
 }
 
 int engine_unregister(struct engine *engine, const char *name)
