@@ -14,6 +14,7 @@
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
+#include <arpa/inet.h>
 #include <stdlib.h>
 #include <string.h>
 #include <getopt.h>
@@ -58,47 +59,84 @@ static inline char *lua_strerror(int lua_err) {
 	}
 }
 
-/*
- * TTY control
+/**
+ * TTY control: process input and free() the buffer.
+ *
+ * For parameters see http://docs.libuv.org/en/v1.x/stream.html#c.uv_read_cb
+ *
+ * - This is just basic read-eval-print; libedit is supported through krsec;
+ * - stream->data represents a bool determining binary output mode (used by kresc);
+ * - binary output: uint32_t length in network order, followed by that many bytes.
  */
-static void tty_read(uv_stream_t *stream, ssize_t nread, const uv_buf_t *buf)
+static void tty_process_input(uv_stream_t *stream, ssize_t nread, const uv_buf_t *buf)
 {
+	char *cmd = buf ? buf->base : NULL; /* To be free()d on return. */
+
 	/* Set output streams */
-	FILE *out = stdout, *outerr = stderr;
+	FILE *out = stdout;
 	uv_os_fd_t stream_fd = 0;
 	if (uv_fileno((uv_handle_t *)stream, &stream_fd)) {
 		uv_close((uv_handle_t *)stream, (uv_close_cb) free);
-		if (buf) {
-			free(buf->base);
-		}
+		free(cmd);
 		return;
 	}
 	if (stream_fd != STDIN_FILENO) {
-		if (nread <= 0) { /* Close if disconnected */
+		if (nread < 0) { /* Close if disconnected */
 			uv_close((uv_handle_t *)stream, (uv_close_cb) free);
-			if (buf) {
-				free(buf->base);
-			}
+		}
+		if (nread <= 0) {
+			free(cmd);
 			return;
 		}
 		uv_os_fd_t dup_fd = dup(stream_fd);
 		if (dup_fd >= 0) {
-			out = outerr = fdopen(dup_fd, "w");
+			out = fdopen(dup_fd, "w");
 		}
 	}
+
 	/* Execute */
-	if (stream && buf && nread > 0) {
-		char *cmd = buf->base;
+	if (stream && cmd && nread > 0) {
+		/* Ensure cmd is 0-terminated */
 		if (cmd[nread - 1] == '\n') {
 			cmd[nread - 1] = '\0';
+		} else {
+			if (nread >= buf->len) { /* only equality should be possible */
+				char *newbuf = realloc(cmd, nread + 1);
+				if (!newbuf)
+					goto finish;
+				cmd = newbuf;
+			}
+			cmd[nread] = '\0';
 		}
-		struct engine *engine = stream->data;
+
+		/* Pseudo-command for switching to "binary output";
+		 * beware: void* <-> bool */
+		bool is_binary = stream->data;
+		if (strcmp(cmd, "__binary") == 0) {
+			stream->data = (void *)(is_binary = true);
+			goto finish;
+		}
+
+		struct engine *engine = ((struct worker_ctx *)stream->loop->data)->engine;
 		lua_State *L = engine->L;
 		int ret = engine_cmd(L, cmd, false);
 		const char *message = "";
 		if (lua_gettop(L) > 0) {
 			message = lua_tostring(L, -1);
 		}
+
+		/* Simpler output in binary mode */
+		if (is_binary) {
+			size_t len_s = strlen(message);
+			if (len_s > UINT32_MAX)
+				goto finish;
+			uint32_t len_n = htonl(len_s);
+			fwrite(&len_n, sizeof(len_n), 1, out);
+			fwrite(message, len_s, 1, out);
+			lua_settop(L, 0);
+			goto finish;
+		}
+
 		/* Log to remote socket if connected */
 		const char *delim = g_quiet ? "" : "> ";
 		if (stream_fd != STDIN_FILENO) {
@@ -118,15 +156,13 @@ static void tty_read(uv_stream_t *stream, ssize_t nread, const uv_buf_t *buf)
 		fprintf(fp_out, "%s", delim);
 		lua_settop(L, 0);
 	}
+finish:
 	fflush(out);
-	if (buf) {
-		free(buf->base);
-	}
+	free(cmd);
 	/* Close if redirected */
 	if (stream_fd != STDIN_FILENO) {
-		fclose(out); /* outerr is the same */
+		fclose(out);
 	}
-
 }
 
 static void tty_alloc(uv_handle_t *handle, size_t suggested, uv_buf_t *buf) {
@@ -143,8 +179,8 @@ static void tty_accept(uv_stream_t *master, int status)
 			free(client);
 			return;
 		 }
-		 client->data = master->data;
-		 uv_read_start((uv_stream_t *)client, tty_alloc, tty_read);
+		 client->data = 0;
+		 uv_read_start((uv_stream_t *)client, tty_alloc, tty_process_input);
 		 /* Write command line */
 		 if (!g_quiet) {
 		 	uv_buf_t buf = { "> ", 2 };
@@ -250,9 +286,13 @@ static void signal_handler(uv_signal_t *handle, int signum)
 	uv_signal_stop(handle);
 }
 
+/** Split away port from the address. */
 static const char *set_addr(char *addr, int *port)
 {
-	char *p = strchr(addr, '#');
+	char *p = strchr(addr, '@');
+	if (!p) {
+		p = strchr(addr, '#');
+	}
 	if (p) {
 		*port = atoi(p + 1);
 		*p = '\0';
@@ -301,15 +341,16 @@ static void help(int argc, char *argv[])
 {
 	printf("Usage: %s [parameters] [rundir]\n", argv[0]);
 	printf("\nParameters:\n"
-	       " -a, --addr=[addr]    Server address (default: localhost#53).\n"
-	       " -t, --tls=[addr]     Server address for TLS (default: off).\n"
-	       " -S, --fd=[fd]        Listen on given fd (handed out by supervisor).\n"
-	       " -T, --tlsfd=[fd]     Listen using TLS on given fd (handed out by supervisor).\n"
-	       " -c, --config=[path]  Config file path (relative to [rundir]) (default: config).\n"
-	       " -k, --keyfile=[path] File containing trust anchors (DS or DNSKEY).\n"
-	       " -f, --forks=N        Start N forks sharing the configuration.\n"
-	       " -q, --quiet          Quiet output, no prompt in interactive mode.\n"
-	       " -v, --verbose        Run in verbose mode."
+	       " -a, --addr=[addr]      Server address (default: localhost@53).\n"
+	       " -t, --tls=[addr]       Server address for TLS (default: off).\n"
+	       " -S, --fd=[fd]          Listen on given fd (handed out by supervisor).\n"
+	       " -T, --tlsfd=[fd]       Listen using TLS on given fd (handed out by supervisor).\n"
+	       " -c, --config=[path]    Config file path (relative to [rundir]) (default: config).\n"
+	       " -k, --keyfile=[path]   File containing trust anchors (DS or DNSKEY).\n"
+	       " -m, --moduledir=[path] Override the default module path (" MODULEDIR ").\n"
+	       " -f, --forks=N          Start N forks sharing the configuration.\n"
+	       " -q, --quiet            Quiet output, no prompt in interactive mode.\n"
+	       " -v, --verbose          Run in verbose mode."
 #ifdef NOVERBOSELOG
 	           " (Recompile without -DNOVERBOSELOG to activate.)"
 #endif
@@ -326,13 +367,13 @@ static int run_worker(uv_loop_t *loop, struct engine *engine, fd_array_t *ipc_se
 	auto_free char *sock_file = NULL;
 	uv_pipe_t pipe;
 	uv_pipe_init(loop, &pipe, 0);
-	pipe.data = engine;
+	pipe.data = 0;
 	if (g_interactive) {
 		if (!g_quiet)
 			printf("[system] interactive mode\n> ");
 		fflush(stdout);
 		uv_pipe_open(&pipe, 0);
-		uv_read_start((uv_stream_t*) &pipe, tty_alloc, tty_read);
+		uv_read_start((uv_stream_t*) &pipe, tty_alloc, tty_process_input);
 	} else {
 		int pipe_ret = -1;
 		if (control_fd != -1) {
@@ -371,13 +412,15 @@ static int run_worker(uv_loop_t *loop, struct engine *engine, fd_array_t *ipc_se
 	return kr_ok();
 }
 
-void free_sd_socket_names(char **socket_names, int count)
+#ifdef HAS_SYSTEMD
+static void free_sd_socket_names(char **socket_names, int count)
 {
 	for (int i = 0; i < count; i++) {
 		free(socket_names[i]);
 	}
 	free(socket_names);
 }
+#endif
 
 int main(int argc, char **argv)
 {
@@ -391,6 +434,7 @@ int main(int argc, char **argv)
 	array_t(int) tls_fd_set;
 	array_init(tls_fd_set);
 	char *keyfile = NULL;
+	char *moduledir = MODULEDIR;
 	const char *config = NULL;
 	int control_fd = -1;
 
@@ -404,13 +448,14 @@ int main(int argc, char **argv)
 		{"config", required_argument, 0, 'c'},
 		{"keyfile",required_argument, 0, 'k'},
 		{"forks",required_argument,   0, 'f'},
+		{"moduledir", required_argument, 0, 'm'},
 		{"verbose",    no_argument,   0, 'v'},
 		{"quiet",      no_argument,   0, 'q'},
 		{"version",   no_argument,    0, 'V'},
 		{"help",      no_argument,    0, 'h'},
 		{0, 0, 0, 0}
 	};
-	while ((c = getopt_long(argc, argv, "a:t:S:T:c:f:k:vqVh", opts, &li)) != -1) {
+	while ((c = getopt_long(argc, argv, "a:t:S:T:c:f:m:k:vqVh", opts, &li)) != -1) {
 		switch (c)
 		{
 		case 'a':
@@ -440,6 +485,9 @@ int main(int argc, char **argv)
 		case 'k':
 			keyfile = optarg;
 			break;
+		case 'm':
+			moduledir = optarg;
+			break;
 		case 'v':
 			kr_verbose_set(true);
 #ifdef NOVERBOSELOG
@@ -500,12 +548,16 @@ int main(int argc, char **argv)
 			kr_log_error("[system] rundir '%s': %s\n", rundir, strerror(errno));
 			return EXIT_FAILURE;
 		}
-		if(config && strcmp(config, "-") != 0 && access(config, R_OK) != 0) {
-			kr_log_error("[system] rundir '%s'\n", rundir);
-			kr_log_error("[system] config '%s': %s\n", config, strerror(errno));
-			return EXIT_FAILURE;
-		}
 	}
+
+	if (config && strcmp(config, "-") != 0 && access(config, R_OK) != 0) {
+		kr_log_error("[system] config '%s': %s\n", config, strerror(errno));
+		return EXIT_FAILURE;
+	}
+	if (!config && access("config", R_OK) == 0) {
+		config = "config";
+	}
+
 #ifndef CAN_FORK_EARLY
 	/* Forking is currently broken with libuv. We need libuv to bind to
 	 * sockets etc. before forking, but at the same time can't touch it before
@@ -570,7 +622,7 @@ int main(int argc, char **argv)
 			const char *addr = set_addr(addr_set.at[i], &port);
 			ret = network_listen(&engine.net, addr, (uint16_t)port, NET_UDP|NET_TCP);
 			if (ret != 0) {
-				kr_log_error("[system] bind to '%s#%d' %s\n", addr, port, kr_strerror(ret));
+				kr_log_error("[system] bind to '%s@%d' %s\n", addr, port, kr_strerror(ret));
 				ret = EXIT_FAILURE;
 				break;
 			}
@@ -583,7 +635,7 @@ int main(int argc, char **argv)
 			const char *addr = set_addr(tls_set.at[i], &port);
 			ret = network_listen(&engine.net, addr, (uint16_t)port, NET_TCP|NET_TLS);
 			if (ret != 0) {
-				kr_log_error("[system] bind to '%s#%d' (TLS) %s\n", addr, port, kr_strerror(ret));
+				kr_log_error("[system] bind to '%s@%d' (TLS) %s\n", addr, port, kr_strerror(ret));
 				ret = EXIT_FAILURE;
 				break;
 			}
@@ -594,6 +646,8 @@ int main(int argc, char **argv)
 		goto cleanup;
 	}
 
+	engine_set_moduledir(&engine, moduledir);
+	
 	/* Block signals. */
 	uv_loop_t *loop = uv_default_loop();
 	uv_signal_t sigint, sigterm;
@@ -605,7 +659,7 @@ int main(int argc, char **argv)
 	worker->loop = loop;
 	loop->data = worker;
 
-	ret = engine_start(&engine, config ? config : "config");
+	ret = engine_start(&engine, config);
 	if (ret != 0) {
 		ret = EXIT_FAILURE;
 		goto cleanup;
@@ -680,9 +734,14 @@ int main(int argc, char **argv)
 			ret =  EXIT_FAILURE;
 			goto cleanup;
 		}
-		int lua_ret = 0;
-		if ((lua_ret = engine_cmd(engine.L, cmd, false)) != 0) {
-			kr_log_error("[ ta ] keyfile '%s': failed to load (%s)\n", keyfile_path, lua_strerror(lua_ret));
+		int lua_ret = engine_cmd(engine.L, cmd, false);
+		if (lua_ret != 0) {
+			if (lua_gettop(engine.L) > 0) {
+				kr_log_error("%s", lua_tostring(engine.L, -1));
+			} else {
+				kr_log_error("[ ta ] keyfile '%s': failed to load (%s)\n",
+						keyfile_path, lua_strerror(lua_ret));
+			}
 			ret = EXIT_FAILURE;
 			goto cleanup;
 		}
