@@ -21,6 +21,9 @@
 #include "lib/defines.h"
 #include "contrib/ucw/config.h" /*uint*/
 
+/** When knot_pkt is passed from cache without ->wire, this is the ->size. */
+static const size_t PKT_SIZE_NOWIRE = -1;
+
 /** Cache entry tag */
 enum kr_cache_tag {
 	KR_CACHE_RR   = 'R',
@@ -29,31 +32,20 @@ enum kr_cache_tag {
 	KR_CACHE_USER = 0x80
 };
 
-/**
- * Cache entry rank.
- * @note Be careful about chosen cache rank nominal values.
- * - AUTH must be > than NONAUTH
- * - AUTH INSECURE must be > than AUTH (because it attempted validation)
- * - NONAUTH SECURE must be > than AUTH (because it's valid)
- *
- * Ref: https://tools.ietf.org/html/rfc2181#section-5.4.1
- */
-enum kr_cache_rank {
-	KR_RANK_BAD       = 0,  /* BAD cache, do not use. */
-	KR_RANK_INSECURE  = 1,  /* Entry is DNSSEC insecure (e.g. RRSIG not exists). */
-	KR_RANK_EXTRA     = 4,  /* Entry from additional section. */
-	KR_RANK_NONAUTH   = 8,  /* Entry from authority section (i.e. parent-side). */
-	KR_RANK_AUTH      = 16, /* Entry from answer (authoritative data). */
-	KR_RANK_SECURE    = 32, /* Entry is DNSSEC valid (e.g. RRSIG exists). */
-	/* @note Rank must not exceed 6 bits */
-};
-
 /** Cache entry flags */
 enum kr_cache_flag {
 	KR_CACHE_FLAG_NONE	  = 0,
 	KR_CACHE_FLAG_WCARD_PROOF = 1, /* Entry contains either packet with wildcard
 	                                * answer either record for which wildcard
 	                                * expansion proof is needed */
+	KR_CACHE_FLAG_OPTOUT	  = 2, /* Entry contains secured packet containing a
+					* closest  encloser proof in which the NSEC3 RR
+					* that covers the "next closer" name
+					* has the Opt-Out bit set
+					*/
+	KR_CACHE_FLAG_NODS	  = 4, /* Entry contains NS rrset
+					* for which DS nonexistence is proven.
+					*/
 };
 
 
