@@ -14,6 +14,19 @@
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
+/** @file iterate.c
+ *
+ * This builtin module is mainly active in the consume phase.
+ * Primary responsibilities:
+ *  - Classify the packet as auth/nonauth and change its AA flag accordingly.
+ *  - Pick interesting RRs to kr_request::answ_selected and ::auth_selected,
+ *    NEW: and classify their rank, except for validation status.
+ *  - Update kr_query::zone_cut (in case of referral).
+ *  - Interpret CNAMEs.
+ *  - Prepare the followup query - either inline or as another kr_query
+ *    (CNAME jumps create a new "sibling" query).
+ */
+
 #include <sys/time.h>
 #include <assert.h>
 #include <arpa/inet.h>
@@ -138,16 +151,29 @@ static int update_nsaddr(const knot_rrset_t *rr, struct kr_query *query)
 {
 	if (rr->type == KNOT_RRTYPE_A || rr->type == KNOT_RRTYPE_AAAA) {
 		const knot_rdata_t *rdata = rr->rrs.data;
+		const void *addr = knot_rdata_data(rdata);
+		const int addr_len = knot_rdata_rdlen(rdata);
+		char name_str[KNOT_DNAME_MAXLEN];
+		char addr_str[INET6_ADDRSTRLEN];
+		WITH_VERBOSE {
+			const int af = (addr_len == sizeof(struct in_addr)) ?
+				       AF_INET : AF_INET6;
+			knot_dname_to_str(name_str, rr->owner, sizeof(name_str));
+			inet_ntop(af, addr, addr_str, sizeof(addr_str));
+		}
 		if (!(query->flags & QUERY_ALLOW_LOCAL) &&
-			!is_valid_addr(knot_rdata_data(rdata), knot_rdata_rdlen(rdata))) {
+			!is_valid_addr(addr, addr_len)) {
+			QVERBOSE_MSG(query, "<= ignoring invalid glue for "
+				     "'%s': '%s'\n", name_str, addr_str);
 			return KR_STATE_CONSUME; /* Ignore invalid addresses */
 		}
 		int ret = kr_zonecut_add(&query->zone_cut, rr->owner, rdata);
 		if (ret != 0) {
 			return KR_STATE_FAIL;
 		}
+		QVERBOSE_MSG(query, "<= using glue for "
+			     "'%s': '%s'\n", name_str, addr_str);
 	}
-
 	return KR_STATE_CONSUME;
 }
 
@@ -165,19 +191,15 @@ static void fetch_glue(knot_pkt_t *pkt, const knot_dname_t *ns, struct kr_reques
 			    (rr->type != KNOT_RRTYPE_AAAA)) {
 				continue;
 			}
-			(void) update_nsaddr(rr, req->current_query);
-			WITH_VERBOSE {
-				char name_str[KNOT_DNAME_MAXLEN];
-				char addr_str[INET6_ADDRSTRLEN];
-				const void *addr = knot_rdata_data(rr->rrs.data);
-				const int addr_len = knot_rdata_rdlen(rr->rrs.data);
-				const int af = (addr_len == sizeof(struct in_addr)) ?
-					       AF_INET : AF_INET6;
-				knot_dname_to_str(name_str, ns, sizeof(name_str));
-				inet_ntop(af, addr, addr_str, sizeof(addr_str));
-				VERBOSE_MSG("<= using glue for '%s': '%s'\n",
-					    name_str, addr_str);
+			if ((rr->type == KNOT_RRTYPE_A) &&
+			    (req->ctx->options & QUERY_NO_IPV4)) {
+				continue;
 			}
+			if ((rr->type == KNOT_RRTYPE_AAAA) &&
+			    (req->ctx->options & QUERY_NO_IPV6)) {
+				continue;
+			}
+			(void) update_nsaddr(rr, req->current_query);
 		}
 	}
 }
@@ -209,8 +231,10 @@ static int update_cut(knot_pkt_t *pkt, const knot_rrset_t *rr,
 	int state = KR_STATE_CONSUME;
 
 	/* New authority MUST be at/below the authority of the current cut;
+	 * also qname must be below new authority;
 	 * otherwise it's a possible cache injection attempt. */
-	if (!knot_dname_in(current_cut, rr->owner)) {
+	if (!knot_dname_in(current_cut, rr->owner) ||
+	    !knot_dname_in(rr->owner, qry->sname)) {
 		VERBOSE_MSG("<= authority: ns outside bailiwick\n");
 #ifdef STRICT_MODE
 		return KR_STATE_FAIL;
@@ -267,12 +291,44 @@ static int update_cut(knot_pkt_t *pkt, const knot_rrset_t *rr,
 	return state;
 }
 
+/** Compute rank appropriate for RRs present in the packet.
+ * @param answer whether the RR is from answer or authority section */
+static uint8_t get_initial_rank(const knot_rrset_t *rr, const struct kr_query *qry,
+				const bool answer, const bool is_referral)
+{
+	/* For RRSIGs, ensure the KR_RANK_AUTH flag corresponds to the signed RR. */
+	uint16_t type = kr_rrset_type_maysig(rr);
+
+	if (qry->flags & QUERY_CACHED) {
+		return rr->additional ? *(uint8_t *)rr->additional : KR_RANK_OMIT;
+		/* ^^ Current use case for "cached" RRs without rank: hints module. */
+	}
+	if (answer || type == KNOT_RRTYPE_DS
+	    || type == KNOT_RRTYPE_NSEC || type == KNOT_RRTYPE_NSEC3) {
+		return KR_RANK_INITIAL | KR_RANK_AUTH;
+	}
+	if (type == KNOT_RRTYPE_NS) {
+		/* Some servers add extra NS RRset, which allows us to refresh
+		 * cache "for free", potentially speeding up zone cut lookups
+		 * in future.  Still, it might theoretically cause some problems:
+		 * https://mailarchive.ietf.org/arch/msg/dnsop/CYjPDlwtpxzdQV_qycB-WfnW6CI
+		 */
+		if (!is_referral && knot_dname_is_equal(qry->zone_cut.name, rr->owner)) {
+			return KR_RANK_INITIAL | KR_RANK_AUTH;
+		} else {
+			return KR_RANK_OMIT;
+		}
+	}
+
+	return KR_RANK_INITIAL;
+	/* TODO: this classifier of authoritativity may not be perfect yet. */
+}
+
 static int pick_authority(knot_pkt_t *pkt, struct kr_request *req, bool to_wire)
 {
 	struct kr_query *qry = req->current_query;
 	const knot_pktsection_t *ns = knot_pkt_section(pkt, KNOT_AUTHORITY);
-	uint8_t rank = !(qry->flags & QUERY_DNSSEC_WANT) || (qry->flags & QUERY_CACHED) ?
-			KR_VLDRANK_SECURE : KR_VLDRANK_INITIAL;
+
 	const knot_dname_t *zonecut_name = qry->zone_cut.name;
 	bool referral = !knot_wire_get_aa(pkt->wire);
 	if (referral) {
@@ -287,6 +343,7 @@ static int pick_authority(knot_pkt_t *pkt, struct kr_request *req, bool to_wire)
 		if (!knot_dname_in(zonecut_name, rr->owner)) {
 			continue;
 		}
+		uint8_t rank = get_initial_rank(rr, qry, false, referral);
 		int ret = kr_ranked_rrarray_add(&req->auth_selected, rr,
 						rank, to_wire, qry->uid, &req->pool);
 		if (ret != kr_ok()) {
@@ -303,6 +360,10 @@ static int process_authority(knot_pkt_t *pkt, struct kr_request *req)
 	assert(!(qry->flags & QUERY_STUB));
 
 	int result = KR_STATE_CONSUME;
+	if (qry->flags & QUERY_FORWARD) {
+		return result;
+	}
+
 	const knot_pktsection_t *ns = knot_pkt_section(pkt, KNOT_AUTHORITY);
 
 #ifdef STRICT_MODE
@@ -319,14 +380,23 @@ static int process_authority(knot_pkt_t *pkt, struct kr_request *req)
 		if (rr->type == KNOT_RRTYPE_CNAME) {
 			return KR_STATE_CONSUME;
 		}
+		/* Work around for these NSs which are authoritative both for
+		 * parent and child and mixes data from both zones in single answer */
+		if (knot_wire_get_aa(pkt->wire) &&
+		    (rr->type == qry->stype) &&
+		    (knot_dname_is_equal(rr->owner, qry->sname))) {
+			return KR_STATE_CONSUME;
+		}
 	}
 #endif
 	/* Remember current bailiwick for NS processing. */
 	const knot_dname_t *current_zone_cut = qry->zone_cut.name;
+	bool ns_record_exists = false;
 	/* Update zone cut information. */
 	for (unsigned i = 0; i < ns->count; ++i) {
 		const knot_rrset_t *rr = knot_pkt_rr(ns, i);
 		if (rr->type == KNOT_RRTYPE_NS) {
+			ns_record_exists = true;
 			int state = update_cut(pkt, rr, req, current_zone_cut);
 			switch(state) {
 			case KR_STATE_DONE: result = state; break;
@@ -339,9 +409,13 @@ static int process_authority(knot_pkt_t *pkt, struct kr_request *req)
 		}
 	}
 
+
 	if ((qry->flags & QUERY_DNSSEC_WANT) && (result == KR_STATE_CONSUME)) {
-		if (knot_wire_get_aa(pkt->wire) == 0 && knot_wire_get_ancount(pkt->wire) == 0) {
-			/* Prevent from validating as an autoritative answer */
+		if (knot_wire_get_aa(pkt->wire) == 0 &&
+		    knot_wire_get_ancount(pkt->wire) == 0 &&
+		    ns_record_exists) {
+			/* Unhelpful referral
+			   Prevent from validating as an authoritative answer */
 			result = KR_STATE_DONE;
 		}
 	}
@@ -358,18 +432,6 @@ static void finalize_answer(knot_pkt_t *pkt, struct kr_query *qry, struct kr_req
 	knot_wire_set_rcode(answer->wire, knot_wire_get_rcode(pkt->wire));
 }
 
-static bool is_rrsig_type_covered(const knot_rrset_t *rr, uint16_t type)
-{
-	bool ret = false;
-	for (unsigned i = 0; i < rr->rrs.rr_count; ++i) {
-		if (knot_rrsig_type_covered(&rr->rrs, i) == type) {
-			ret = true;
-			break;
-		}
-	}
-	return ret;
-}
-
 static int unroll_cname(knot_pkt_t *pkt, struct kr_request *req, bool referral, const knot_dname_t **cname_ret)
 {
 	struct kr_query *query = req->current_query;
@@ -379,23 +441,40 @@ static int unroll_cname(knot_pkt_t *pkt, struct kr_request *req, bool referral,
 	const knot_dname_t *cname = NULL;
 	const knot_dname_t *pending_cname = query->sname;
 	unsigned cname_chain_len = 0;
-	uint8_t rank = !(query->flags & QUERY_DNSSEC_WANT) || (query->flags & QUERY_CACHED) ?
-			KR_VLDRANK_SECURE : KR_VLDRANK_INITIAL;
 	bool is_final = (query->parent == NULL);
-	bool can_follow = false;
+	uint32_t iter_count = 0;
 	bool strict_mode = (query->flags & QUERY_STRICT);
 	do {
 		/* CNAME was found at previous iteration, but records may not follow the correct order.
 		 * Try to find records for pending_cname owner from section start. */
 		cname = pending_cname;
 		pending_cname = NULL;
-		/* If not secure, always follow cname chain. */
-		can_follow = !(query->flags & QUERY_DNSSEC_WANT);
+		const int cname_labels = knot_dname_labels(cname, NULL);
 		for (unsigned i = 0; i < an->count; ++i) {
 			const knot_rrset_t *rr = knot_pkt_rr(an, i);
-			if (!knot_dname_is_equal(rr->owner, cname)) {
+
+			/* Skip the RR if its owner+type doesn't interest us. */
+			const uint16_t type = kr_rrset_type_maysig(rr);
+			const bool type_OK = rr->type == query->stype || type == query->stype
+				|| type == KNOT_RRTYPE_CNAME || type == KNOT_RRTYPE_DNAME;
+				/* TODO: actually handle DNAMEs */
+			if (!type_OK || !knot_dname_is_equal(rr->owner, cname)) {
 				continue;
 			}
+
+			if (rr->type == KNOT_RRTYPE_RRSIG) {
+				int rrsig_labels = knot_rrsig_labels(&rr->rrs, 0);
+				if (rrsig_labels > cname_labels) {
+					/* clearly wrong RRSIG, don't pick it.
+					 * don't fail immediately,
+					 * let validator work. */
+					continue;
+				}
+				if (rrsig_labels < cname_labels) {
+					query->flags |= QUERY_DNSSEC_WEXPAND;
+				}
+			}
+
 			/* Process records matching current SNAME */
 			int state = KR_STATE_FAIL;
 			bool to_wire = false;
@@ -408,17 +487,12 @@ static int unroll_cname(knot_pkt_t *pkt, struct kr_request *req, bool referral,
 					return state;
 				}
 			}
+			uint8_t rank = get_initial_rank(rr, query, true, referral);
 			state = kr_ranked_rrarray_add(&req->answ_selected, rr,
 						      rank, to_wire, query->uid, &req->pool);
 			if (state != kr_ok()) {
 				return KR_STATE_FAIL;
 			}
-			/* can_follow is false, therefore QUERY_DNSSEC_WANT flag is set.
-			 * Follow cname chain only if rrsig exists. */
-			if (!can_follow && rr->type == KNOT_RRTYPE_RRSIG &&
-			    is_rrsig_type_covered(rr, KNOT_RRTYPE_CNAME)) {
-				can_follow = true;
-			}
 			/* Jump to next CNAME target */
 			if ((query->stype == KNOT_RRTYPE_CNAME) || (rr->type != KNOT_RRTYPE_CNAME)) {
 				continue;
@@ -449,7 +523,6 @@ static int unroll_cname(knot_pkt_t *pkt, struct kr_request *req, bool referral,
 		}
 		/* try to unroll cname only within current zone */
 		const int pending_labels = knot_dname_labels(pending_cname, NULL);
-		const int cname_labels = knot_dname_labels(cname, NULL);
 		if (pending_labels != cname_labels) {
 			cname = pending_cname;
 			break;
@@ -459,7 +532,11 @@ static int unroll_cname(knot_pkt_t *pkt, struct kr_request *req, bool referral,
 			cname = pending_cname;
 			break;
 		}
-	} while (can_follow);
+	} while (++iter_count < KR_CNAME_CHAIN_LIMIT);
+	if (iter_count >= KR_CNAME_CHAIN_LIMIT) {
+		VERBOSE_MSG("<= too long cname chain\n");
+		return KR_STATE_FAIL;
+	}
 	*cname_ret = cname;
 	return kr_ok();
 }
@@ -471,6 +548,18 @@ static int process_referral_answer(knot_pkt_t *pkt, struct kr_request *req)
 	if (state != kr_ok()) {
 		return KR_STATE_FAIL;
 	}
+	struct kr_query *query = req->current_query;
+	if (!(query->flags & QUERY_CACHED)) {
+		/* If not cached (i.e. got from upstream)
+		 * make sure that this is not an authoritative answer
+		 * (even with AA=1) for other layers.
+		 * There can be answers with AA=1,
+		 * empty answer section and NS in authority.
+		 * Clearing of AA prevents them from
+		 * caching in the packet cache.
+		 * If packet already cached, don't touch him. */
+		knot_wire_clear_aa(pkt->wire);
+	}
 	state = pick_authority(pkt, req, false);
 	return state == kr_ok() ? KR_STATE_DONE : KR_STATE_FAIL;
 }
@@ -533,7 +622,8 @@ static int process_answer(knot_pkt_t *pkt, struct kr_request *req)
 
 	/* This answer didn't improve resolution chain, therefore must be authoritative (relaxed to negative). */
 	if (!is_authoritative(pkt, query)) {
-		if (pkt_class & (PKT_NXDOMAIN|PKT_NODATA)) {
+		if (!(query->flags & QUERY_FORWARD) &&
+		    pkt_class & (PKT_NXDOMAIN|PKT_NODATA)) {
 			VERBOSE_MSG("<= lame response: non-auth sent negative response\n");
 			return KR_STATE_FAIL;
 		}
@@ -558,13 +648,18 @@ static int process_answer(knot_pkt_t *pkt, struct kr_request *req)
 			if (state != kr_ok()) {
 				return state;
 			}
+		} else if ((query->flags & QUERY_FORWARD) &&
+			   ((query->stype == KNOT_RRTYPE_DS) ||
+			    (query->stype == KNOT_RRTYPE_NS))) {
+			/* CNAME'ed answer for DS or NS subquery.
+			 * Treat it as proof of zonecut nonexistance. */
+			return KR_STATE_DONE;
 		}
 		VERBOSE_MSG("<= cname chain, following\n");
-		/* Check if the same query was already resolved */
-		for (int i = 0; i < req->rplan.resolved.len; ++i) {
-			struct kr_query * q = req->rplan.resolved.at[i];
-			if (q->parent == query->parent &&
-			    q->sclass == query->sclass &&
+		/* Check if the same query was followed in the same CNAME chain. */
+		for (const struct kr_query *q = query->cname_parent; q != NULL;
+				q = q->cname_parent) {
+			if (q->sclass == query->sclass &&
 			    q->stype == query->stype   &&
 			    knot_dname_is_equal(q->sname, cname)) {
 				VERBOSE_MSG("<= cname chain loop\n");
@@ -576,20 +671,32 @@ static int process_answer(knot_pkt_t *pkt, struct kr_request *req)
 			return KR_STATE_FAIL;
 		}
 		next->flags |= QUERY_AWAIT_CUT;
-		if (query->flags & QUERY_DNSSEC_INSECURE) {
-			next->flags &= ~QUERY_DNSSEC_WANT;
-			next->flags |= QUERY_DNSSEC_INSECURE;
-		} else if (kr_ta_covers(&req->ctx->trust_anchors, cname) &&
-		    !kr_ta_covers(&req->ctx->negative_anchors, cname)) {
-			/* Want DNSSEC if it's posible to secure
-			 * this name (e.g. is covered by any TA) */
-			next->flags |= QUERY_DNSSEC_WANT;
+		if (query->flags & QUERY_FORWARD) {
+			next->forward_flags |= QUERY_CNAME;
+			if (query->parent == NULL) {
+				state = kr_nsrep_copy_set(&next->ns, &query->ns);
+				if (state != kr_ok()) {
+					return KR_STATE_FAIL;
+				}
+			}
 		}
-		state = pick_authority(pkt, req, false);
-		if (state != kr_ok()) {
-			return KR_STATE_FAIL;
+		next->cname_parent = query;
+		/* Want DNSSEC if and only if it's posible to secure
+		 * this name (i.e. iff it is covered by a TA) */
+		if (kr_ta_covers_qry(req->ctx, cname, query->stype)) {
+			next->flags |= QUERY_DNSSEC_WANT;
+		} else {
+			next->flags &= ~QUERY_DNSSEC_WANT;
+		}
+		if (!(query->flags & QUERY_FORWARD) ||
+		    (query->flags & QUERY_DNSSEC_WEXPAND)) {
+			state = pick_authority(pkt, req, false);
+			if (state != kr_ok()) {
+				return KR_STATE_FAIL;
+			}
 		}
 	} else if (!query->parent) {
+		/* Answer for initial query */
 		const bool to_wire = ((pkt_class & (PKT_NXDOMAIN|PKT_NODATA)) != 0);
 		state = pick_authority(pkt, req, to_wire);
 		if (state != kr_ok()) {
@@ -597,9 +704,33 @@ static int process_answer(knot_pkt_t *pkt, struct kr_request *req)
 		}
 		finalize_answer(pkt, query, req);
 	} else {
-		state = pick_authority(pkt, req, false);
-		if (state != kr_ok()) {
-			return KR_STATE_FAIL;
+		/* Answer for sub-query; DS, IP for NS etc.
+		 * It may contains NSEC \ NSEC3 records for
+		 * data non-existence or wc expansion proving.
+		 * If yes, they must be validated by validator.
+		 * If no, authority section is unuseful.
+		 * dnssec\nsec.c & dnssec\nsec3.c use
+		 * rrsets from incoming packet.
+		 * validator uses answer_selected & auth_selected.
+		 * So, if nsec\nsec3 records are present in authority,
+		 * pick_authority() must be called.
+		 * TODO refactor nsec\nsec3 modules to work with
+		 * answer_selected & auth_selected instead of incoming pkt. */
+		bool auth_is_unuseful = true;
+		const knot_pktsection_t *ns = knot_pkt_section(pkt, KNOT_AUTHORITY);
+		for (unsigned i = 0; i < ns->count; ++i) {
+			const knot_rrset_t *rr = knot_pkt_rr(ns, i);
+			if (rr->type == KNOT_RRTYPE_NSEC ||
+			    rr->type == KNOT_RRTYPE_NSEC3) {
+				auth_is_unuseful = false;
+				break;
+			}
+		}
+		if (!auth_is_unuseful) {
+			state = pick_authority(pkt, req, false);
+			if (state != kr_ok()) {
+				return KR_STATE_FAIL;
+			}
 		}
 	}
 	return KR_STATE_DONE;
@@ -615,7 +746,10 @@ static int process_stub(knot_pkt_t *pkt, struct kr_request *req)
 	for (unsigned i = 0; i < an->count; ++i) {
 		const knot_rrset_t *rr = knot_pkt_rr(an, i);
 		int err = kr_ranked_rrarray_add(&req->answ_selected, rr,
-			      KR_VLDRANK_INITIAL, true, query->uid, &req->pool);
+			      KR_RANK_INITIAL | KR_RANK_AUTH, true, query->uid, &req->pool);
+		/* KR_RANK_AUTH: we don't have the records directly from
+		 * an authoritative source, but we do trust the server and it's
+		 * supposed to only send us authoritative records. */
 		if (err != kr_ok()) {
 			return KR_STATE_FAIL;
 		}
@@ -623,7 +757,6 @@ static int process_stub(knot_pkt_t *pkt, struct kr_request *req)
 
 	knot_wire_set_aa(pkt->wire);
 	query->flags |= QUERY_RESOLVED;
-
 	/* Pick authority RRs. */
 	int pkt_class = kr_response_classify(pkt);
 	const bool to_wire = ((pkt_class & (PKT_NXDOMAIN|PKT_NODATA)) != 0);
@@ -690,7 +823,8 @@ int kr_make_query(struct kr_query *query, knot_pkt_t *pkt)
 	}
 
 	/* Query built, expect answer. */
-	query->id = kr_rand_uint(UINT16_MAX);
+	uint32_t rnd = kr_rand_uint(0);
+	query->id = rnd ^ (rnd >> 16); /* cheap way to strengthen unpredictability */
 	knot_wire_set_id(pkt->wire, query->id);
 	pkt->parsed = pkt->size;
 	WITH_VERBOSE {
@@ -809,7 +943,10 @@ static int resolve(kr_layer_t *ctx, knot_pkt_t *pkt)
 		break; /* OK */
 	case KNOT_RCODE_REFUSED:
 	case KNOT_RCODE_SERVFAIL: {
-		if (query->flags & QUERY_STUB) { break; } /* Pass through in stub mode */
+		if (query->flags & (QUERY_STUB | QUERY_FORWARD)) {
+			 /* Pass through in stub mode */
+			break;
+		}
 		VERBOSE_MSG("<= rcode: %s\n", rcode ? rcode->name : "??");
 		query->fails += 1;
 		if (query->fails >= KR_QUERY_NSRETRY_LIMIT) {
