@@ -91,33 +91,36 @@ static int validate_section(kr_rrset_validation_ctx_t *vctx, knot_mm_t *pool)
 	for (ssize_t i = 0; i < vctx->rrs->len; ++i) {
 		ranked_rr_array_entry_t *entry = vctx->rrs->at[i];
 		const knot_rrset_t *rr = entry->rr;
-		if (entry->rank == KR_VLDRANK_SECURE ||
-		    entry->yielded || vctx->qry_uid != entry->qry_uid) {
+
+		if (entry->yielded || vctx->qry_uid != entry->qry_uid) {
 			continue;
 		}
+
+		if (kr_rank_test(entry->rank, KR_RANK_OMIT)
+		    || kr_rank_test(entry->rank, KR_RANK_SECURE)) {
+			continue; /* these are already OK */
+		}
+
 		if (rr->type == KNOT_RRTYPE_RRSIG) {
 			const knot_dname_t *signer_name = knot_rrsig_signer_name(&rr->rrs, 0);
 			if (!knot_dname_is_equal(vctx->zone_name, signer_name)) {
-				entry->rank = KR_VLDRANK_MISMATCH;
+				kr_rank_set(&entry->rank, KR_RANK_MISMATCH);
 				vctx->err_cnt += 1;
 				break;
 			}
-			entry->rank = KR_VLDRANK_SECURE;
-			continue;
-		}
-		if ((rr->type == KNOT_RRTYPE_NS) && (vctx->section_id == KNOT_AUTHORITY)) {
-			entry->rank = KR_VLDRANK_SECURE;
+			kr_rank_set(&entry->rank, KR_RANK_OMIT);
 			continue;
 		}
+
 		validation_result = kr_rrset_validate(vctx, rr);
 		if (validation_result == kr_ok()) {
-			entry->rank = KR_VLDRANK_SECURE;
+			kr_rank_set(&entry->rank, KR_RANK_SECURE);
 		} else if (validation_result == kr_error(ENOENT)) {
 			/* no RRSIGs found */
-			entry->rank = KR_VLDRANK_INSECURE;
+			kr_rank_set(&entry->rank, KR_RANK_MISSING);
 			vctx->err_cnt += 1;
 		} else {
-			entry->rank = KR_VLDRANK_UNKNOWN;
+			kr_rank_set(&entry->rank, KR_RANK_BOGUS);
 			vctx->err_cnt += 1;
 		}
 	}
@@ -166,11 +169,15 @@ static int validate_records(struct kr_request *req, knot_pkt_t *answer, knot_mm_
 	}
 
 	/* Records were validated.
-	 * If there is wildcard expansion in answer, flag the query.
+	 * If there is wildcard expansion in answer,
+	 * or optout - flag the query.
          */
 	if (an_flags & KR_DNSSEC_VFLG_WEXPAND) {
 		qry->flags |= QUERY_DNSSEC_WEXPAND;
 	}
+	if (an_flags & KR_DNSSEC_VFLG_OPTOUT) {
+		qry->flags |= QUERY_DNSSEC_OPTOUT;
+	}
 
 	return ret;
 }
@@ -225,10 +232,12 @@ static int validate_keyset(struct kr_request *req, knot_pkt_t *answer, bool has_
 			return ret;
 		}
 
-		if (vctx.flags & KR_DNSSEC_VFLG_WEXPAND)
-		{
+		if (vctx.flags & KR_DNSSEC_VFLG_WEXPAND) {
 			qry->flags |= QUERY_DNSSEC_WEXPAND;
 		}
+		if (vctx.flags & KR_DNSSEC_VFLG_OPTOUT) {
+			qry->flags |= QUERY_DNSSEC_OPTOUT;
+		}
 
 	}
 	return kr_ok();
@@ -280,8 +289,9 @@ static void mark_insecure_parents(const struct kr_query *qry)
 	}
 }
 
-static int update_parent_keys(struct kr_query *qry, uint16_t answer_type)
+static int update_parent_keys(struct kr_request *req, uint16_t answer_type)
 {
+	struct kr_query *qry = req->current_query;
 	struct kr_query *parent = qry->parent;
 	assert(parent);
 	switch(answer_type) {
@@ -294,8 +304,25 @@ static int update_parent_keys(struct kr_query *qry, uint16_t answer_type)
 		break;
 	case KNOT_RRTYPE_DS:
 		VERBOSE_MSG(qry, "<= parent: updating DS\n");
-		if (qry->flags & QUERY_DNSSEC_INSECURE) { /* DS non-existence proven. */
+		if (qry->flags & (QUERY_DNSSEC_INSECURE)) { /* DS non-existence proven. */
 			mark_insecure_parents(qry);
+		} else if ((qry->flags & (QUERY_DNSSEC_NODS | QUERY_FORWARD)) == QUERY_DNSSEC_NODS) {
+			if (qry->flags & QUERY_DNSSEC_OPTOUT) {
+				mark_insecure_parents(qry);
+			} else {
+				int ret = kr_dnssec_matches_name_and_type(&req->auth_selected, qry->uid,
+									  qry->sname, KNOT_RRTYPE_NS);
+				if (ret == kr_ok()) {
+					mark_insecure_parents(qry);
+				}
+			}
+		} else if ((qry->flags & (QUERY_DNSSEC_NODS | QUERY_FORWARD)) ==
+			   (QUERY_DNSSEC_NODS | QUERY_FORWARD)) {
+			int ret = kr_dnssec_matches_name_and_type(&req->auth_selected, qry->uid,
+								  qry->sname, KNOT_RRTYPE_NS);
+			if (ret == kr_ok()) {
+				mark_insecure_parents(qry);
+			}
 		} else { /* DS existence proven. */
 			parent->zone_cut.trust_anchor = knot_rrset_copy(qry->zone_cut.trust_anchor, parent->zone_cut.pool);
 			if (!parent->zone_cut.trust_anchor) {
@@ -318,10 +345,14 @@ static int update_delegation(struct kr_request *req, struct kr_query *qry, knot_
 	 * otherwise referral is bogus (or an attempted downgrade attack).
 	 */
 
+
 	unsigned section = KNOT_ANSWER;
-	if (!knot_wire_get_aa(answer->wire)) { /* Referral */
+	const bool referral = !knot_wire_get_aa(answer->wire);
+	if (referral) {
 		section = KNOT_AUTHORITY;
-	} else if (knot_pkt_qtype(answer) == KNOT_RRTYPE_DS) { /* Subrequest */
+	} else if (knot_pkt_qtype(answer) == KNOT_RRTYPE_DS &&
+		   !(qry->flags & QUERY_CNAME) &&
+		   (knot_wire_get_rcode(answer->wire) != KNOT_RCODE_NXDOMAIN)) {
 		section = KNOT_ANSWER;
 	} else { /* N/A */
 		return kr_ok();
@@ -334,16 +365,16 @@ static int update_delegation(struct kr_request *req, struct kr_query *qry, knot_
 	if (!new_ds) {
 		/* No DS provided, check for proof of non-existence. */
 		if (!has_nsec3) {
-			if (!knot_wire_get_aa(answer->wire)) {
-				/* Referral, check if it is referral to unsigned, rfc4035 5.2 */
+			if (referral) {
+				/* Check if it is referral to unsigned, rfc4035 5.2 */
 				ret = kr_nsec_ref_to_unsigned(answer);
 			} else {
 				/* No-data answer */
 				ret = kr_nsec_existence_denial(answer, KNOT_AUTHORITY, proved_name, KNOT_RRTYPE_DS);
 			}
 		} else {
-			if (!knot_wire_get_aa(answer->wire)) {
-				/* Referral, check if it is referral to unsigned, rfc5155 8.9 */
+			if (referral) {
+				/* Check if it is referral to unsigned, rfc5155 8.9 */
 				ret = kr_nsec3_ref_to_unsigned(answer);
 			} else {
 				/* No-data answer, QTYPE is DS, rfc5155 8.6 */
@@ -355,8 +386,7 @@ static int update_delegation(struct kr_request *req, struct kr_query *qry, knot_
 			}
 		}
 
-		if (!knot_wire_get_aa(answer->wire) &&
-		    qry->stype != KNOT_RRTYPE_DS &&
+		if (referral && qry->stype != KNOT_RRTYPE_DS &&
 		    ret == DNSSEC_NOT_FOUND) {
 			/* referral,
 			 * qtype is not KNOT_RRTYPE_DS, NSEC\NSEC3 were not found.
@@ -371,10 +401,12 @@ static int update_delegation(struct kr_request *req, struct kr_query *qry, knot_
 			qry->flags |= QUERY_DNSSEC_BOGUS;
 		} else {
 			VERBOSE_MSG(qry, "<= DS doesn't exist, going insecure\n");
-			qry->flags &= ~QUERY_DNSSEC_WANT;
-			qry->flags |= QUERY_DNSSEC_INSECURE;
+			qry->flags |= QUERY_DNSSEC_NODS;
 		}
 		return ret;
+	} else if (qry->flags & QUERY_FORWARD && qry->parent) {
+		struct kr_query *parent = qry->parent;
+		parent->zone_cut.name = knot_dname_copy(qry->sname, parent->zone_cut.pool);
 	}
 
 	/* Extend trust anchor */
@@ -388,7 +420,9 @@ static const knot_dname_t *find_first_signer(ranked_rr_array_t *arr)
 	for (size_t i = 0; i < arr->len; ++i) {
 		ranked_rr_array_entry_t *entry = arr->at[i];
 		const knot_rrset_t *rr = entry->rr;
-		if (entry->yielded || entry->rank != KR_VLDRANK_INITIAL) {
+		if (entry->yielded ||
+		    (!kr_rank_test(entry->rank, KR_RANK_INITIAL) &&
+		    !kr_rank_test(entry->rank, KR_RANK_MISMATCH))) {
 			continue;
 		}
 		if (rr->type == KNOT_RRTYPE_RRSIG) {
@@ -412,7 +446,7 @@ static int rrsig_not_found(kr_layer_t *ctx, const knot_rrset_t *rr)
 	struct kr_request *req = ctx->req;
 	struct kr_query *qry = req->current_query;
 
-	VERBOSE_MSG(qry, ">< no RRSIGs found\n");
+	VERBOSE_MSG(qry, ">< no valid RRSIGs found\n");
 	struct kr_zonecut *cut = &qry->zone_cut;
 	const knot_dname_t *cut_name_start = qry->zone_cut.name;
 	bool use_cut = true;
@@ -445,6 +479,9 @@ static int rrsig_not_found(kr_layer_t *ctx, const knot_rrset_t *rr)
 	} else {
 		next->flags |= QUERY_AWAIT_CUT;
 	}
+	if (qry->flags & QUERY_FORWARD) {
+		next->flags &= ~QUERY_AWAIT_CUT;
+	}
 	next->flags |= QUERY_DNSSEC_WANT;
 	return KR_STATE_YIELD;
 }
@@ -460,13 +497,15 @@ static int check_validation_result(kr_layer_t *ctx, ranked_rr_array_t *arr)
 		if (entry->yielded || entry->qry_uid != qry->uid) {
 			continue;
 		}
-		if (entry->rank == KR_VLDRANK_MISMATCH) {
+		if (kr_rank_test(entry->rank, KR_RANK_MISMATCH)) {
 			invalid_entry = entry;
 			break;
-		} else if (entry->rank == KR_VLDRANK_INSECURE &&
+		} else if (kr_rank_test(entry->rank, KR_RANK_MISSING) &&
 			   !invalid_entry) {
 			invalid_entry = entry;
-		} else if (entry->rank != KR_VLDRANK_SECURE &&
+		} else if (kr_rank_test(entry->rank, KR_RANK_OMIT)) {
+			continue;
+		} else if (!kr_rank_test(entry->rank, KR_RANK_SECURE) &&
 			   !invalid_entry) {
 			invalid_entry = entry;
 		}
@@ -476,7 +515,7 @@ static int check_validation_result(kr_layer_t *ctx, ranked_rr_array_t *arr)
 		return ret;
 	}
 
-	if ((invalid_entry->rank != KR_VLDRANK_SECURE) &&
+	if (!kr_rank_test(invalid_entry->rank, KR_RANK_SECURE) &&
 	    (++(invalid_entry->revalidation_cnt) > MAX_REVALIDATION_CNT)) {
 		VERBOSE_MSG(qry, "<= continuous revalidation, fails\n");
 		qry->flags |= QUERY_DNSSEC_BOGUS;
@@ -484,7 +523,7 @@ static int check_validation_result(kr_layer_t *ctx, ranked_rr_array_t *arr)
 	}
 
 	const knot_rrset_t *rr = invalid_entry->rr;
-	if (invalid_entry->rank == KR_VLDRANK_MISMATCH) {
+	if (kr_rank_test(invalid_entry->rank, KR_RANK_MISMATCH)) {
 		const knot_dname_t *signer_name = knot_rrsig_signer_name(&rr->rrs, 0);
 		if (knot_dname_is_sub(signer_name, qry->zone_cut.name)) {
 			qry->zone_cut.name = knot_dname_copy(signer_name, &req->pool);
@@ -499,9 +538,9 @@ static int check_validation_result(kr_layer_t *ctx, ranked_rr_array_t *arr)
 		}
 		VERBOSE_MSG(qry, ">< cut changed (new signer), needs revalidation\n");
 		ret = KR_STATE_YIELD;
-	} else if (invalid_entry->rank == KR_VLDRANK_INSECURE) {
+	} else if (kr_rank_test(invalid_entry->rank, KR_RANK_MISSING)) {
 		ret = rrsig_not_found(ctx, rr);
-	} else if (invalid_entry->rank != KR_VLDRANK_SECURE) {
+	} else if (!kr_rank_test(invalid_entry->rank, KR_RANK_SECURE)) {
 		qry->flags |= QUERY_DNSSEC_BOGUS;
 		ret = KR_STATE_FAIL;
 	}
@@ -509,6 +548,95 @@ static int check_validation_result(kr_layer_t *ctx, ranked_rr_array_t *arr)
 	return ret;
 }
 
+static bool check_empty_answer(kr_layer_t *ctx, knot_pkt_t *pkt)
+{
+	struct kr_request *req	= ctx->req;
+	struct kr_query *qry	= req->current_query;
+	ranked_rr_array_t *arr	= &req->answ_selected;
+	size_t num_entries = 0;
+	for (size_t i = 0; i < arr->len; ++i) {
+		ranked_rr_array_entry_t *entry = arr->at[i];
+		const knot_rrset_t *rr = entry->rr;
+		if (rr->type == KNOT_RRTYPE_RRSIG && qry->stype != KNOT_RRTYPE_RRSIG) {
+			continue;
+		}
+		if (entry->qry_uid == qry->uid) {
+			++num_entries;
+		}
+	}
+	const knot_pktsection_t *an = knot_pkt_section(pkt, KNOT_ANSWER);
+	return ((an->count != 0) && (num_entries == 0)) ? false : true;
+}
+
+static int unsigned_forward(kr_layer_t *ctx, knot_pkt_t *pkt)
+{
+	struct kr_request *req = ctx->req;
+	struct kr_query *qry = req->current_query;
+	const uint16_t qtype = knot_pkt_qtype(pkt);
+	const uint8_t pkt_rcode = knot_wire_get_rcode(pkt->wire);
+	bool nods = false;
+	bool ns_exist = true;
+	for (int i = 0; i < req->rplan.resolved.len; ++i) {
+		struct kr_query *q = req->rplan.resolved.at[i];
+		if (q->sclass == qry->sclass &&
+		    q->stype == KNOT_RRTYPE_DS &&
+		    knot_dname_is_equal(q->sname, qry->sname)) {
+			nods = true;
+			if (!(q->flags & QUERY_DNSSEC_OPTOUT)) {
+				int ret = kr_dnssec_matches_name_and_type(&req->auth_selected, q->uid,
+									  qry->sname, KNOT_RRTYPE_NS);
+				ns_exist = (ret == kr_ok());
+			}
+		}
+	}
+
+	if (nods && ns_exist && qtype == KNOT_RRTYPE_NS) {
+		qry->flags &= ~QUERY_DNSSEC_WANT;
+		qry->flags |= QUERY_DNSSEC_INSECURE;
+		if (qry->forward_flags & QUERY_CNAME) {
+			assert(qry->cname_parent);
+			qry->cname_parent->flags &= ~QUERY_DNSSEC_WANT;
+			qry->cname_parent->flags |= QUERY_DNSSEC_INSECURE;
+		} else if (pkt_rcode == KNOT_RCODE_NOERROR && qry->parent != NULL) {
+			const knot_pktsection_t *sec = knot_pkt_section(pkt, KNOT_ANSWER);
+			const knot_rrset_t *rr = knot_pkt_rr(sec, 0);
+			if (rr->type == KNOT_RRTYPE_NS) {
+				qry->parent->zone_cut.name = knot_dname_copy(rr->owner, &req->pool);
+				qry->parent->flags &= ~QUERY_DNSSEC_WANT;
+				qry->parent->flags |= QUERY_DNSSEC_INSECURE;
+			}
+		}
+		while (qry->parent) {
+			qry = qry->parent;
+			qry->flags &= ~QUERY_DNSSEC_WANT;
+			qry->flags |= QUERY_DNSSEC_INSECURE;
+			if (qry->forward_flags & QUERY_CNAME) {
+				assert(qry->cname_parent);
+				qry->cname_parent->flags &= ~QUERY_DNSSEC_WANT;
+				qry->cname_parent->flags |= QUERY_DNSSEC_INSECURE;
+			}
+		}
+		return KR_STATE_DONE;
+	}
+
+	if (ctx->state == KR_STATE_YIELD) {
+		return KR_STATE_DONE;
+	}
+
+	if (!nods && qtype != KNOT_RRTYPE_DS) {
+		struct kr_rplan *rplan = &req->rplan;
+		struct kr_query *next = kr_rplan_push(rplan, qry, qry->sname, qry->sclass, KNOT_RRTYPE_DS);
+		if (!next) {
+			return KR_STATE_FAIL;
+		}
+		kr_zonecut_set(&next->zone_cut, qry->zone_cut.name);
+		kr_zonecut_copy_trust(&next->zone_cut, &qry->zone_cut);
+		next->flags |= QUERY_DNSSEC_WANT;
+	}
+
+	return KR_STATE_YIELD;
+}
+
 static int check_signer(kr_layer_t *ctx, knot_pkt_t *pkt)
 {
 	struct kr_request *req = ctx->req;
@@ -518,6 +646,9 @@ static int check_signer(kr_layer_t *ctx, knot_pkt_t *pkt)
 	if (ta_name && (!signer || !knot_dname_is_equal(ta_name, signer))) {
 		/* check all newly added RRSIGs */
 		if (!signer) {
+			if (qry->flags & QUERY_FORWARD) {
+				return unsigned_forward(ctx, pkt);
+			}
 			/* Not a DNSSEC-signed response. */
 			if (ctx->state == KR_STATE_YIELD) {
 				/* Already yielded for revalidation.
@@ -538,8 +669,32 @@ static int check_signer(kr_layer_t *ctx, knot_pkt_t *pkt)
 				qry->zone_cut.name = knot_dname_copy(qname, &req->pool);
 			}
 		} else if (knot_dname_is_sub(signer, qry->zone_cut.name)) {
-			/* Key signer is below current cut, advance and refetch keys. */
-			qry->zone_cut.name = knot_dname_copy(signer, &req->pool);
+			if (!(qry->flags & QUERY_FORWARD)) {
+				/* Key signer is below current cut, advance and refetch keys. */
+				qry->zone_cut.name = knot_dname_copy(signer, &req->pool);
+			} else {
+				/* Check if DS does not exist. */
+				struct kr_query *q = kr_rplan_find_resolved(&req->rplan, NULL,
+									    signer, qry->sclass, KNOT_RRTYPE_DS);
+				if (q && q->flags & QUERY_DNSSEC_NODS) {
+					qry->flags &= ~QUERY_DNSSEC_WANT;
+					qry->flags |= QUERY_DNSSEC_INSECURE;
+					if (qry->parent) {
+						qry->parent->flags &= ~QUERY_DNSSEC_WANT;
+						qry->parent->flags |= QUERY_DNSSEC_INSECURE;
+					}
+				} else if (qry->stype != KNOT_RRTYPE_DS) {
+					struct kr_rplan *rplan = &req->rplan;
+					struct kr_query *next = kr_rplan_push(rplan, qry, qry->sname,
+									      qry->sclass, KNOT_RRTYPE_DS);
+					if (!next) {
+						return KR_STATE_FAIL;
+					}
+					kr_zonecut_set(&next->zone_cut, qry->zone_cut.name);
+					kr_zonecut_copy_trust(&next->zone_cut, &qry->zone_cut);
+					next->flags |= QUERY_DNSSEC_WANT;
+				}
+			}
 		} else if (!knot_dname_is_equal(signer, qry->zone_cut.name)) {
 			/* Key signer is above the current cut, so we can't validate it. This happens when
 			   a server is authoritative for both grandparent, parent and child zone.
@@ -550,18 +705,87 @@ static int check_signer(kr_layer_t *ctx, knot_pkt_t *pkt)
 				qry->flags |= QUERY_AWAIT_CUT;
 			}
 			qry->zone_cut.name = knot_dname_copy(signer, &req->pool);
-		} /* else zone cut matches, but DS/DNSKEY doesn't => refetch. */
+		}
+
+		/* zone cut matches, but DS/DNSKEY doesn't => refetch. */
 		VERBOSE_MSG(qry, ">< cut changed, needs revalidation\n");
-		return KR_STATE_YIELD;
+		if ((qry->flags & QUERY_FORWARD) && qry->stype != KNOT_RRTYPE_DS) {
+			struct kr_rplan *rplan = &req->rplan;
+			struct kr_query *next = kr_rplan_push(rplan, qry, signer,
+							qry->sclass, KNOT_RRTYPE_DS);
+			if (!next) {
+				return KR_STATE_FAIL;
+			}
+			kr_zonecut_set(&next->zone_cut, qry->zone_cut.name);
+			kr_zonecut_copy_trust(&next->zone_cut, &qry->zone_cut);
+			next->flags |= QUERY_DNSSEC_WANT;
+			return KR_STATE_YIELD;
+		}
+		if (!(qry->flags & QUERY_FORWARD)) {
+			return KR_STATE_YIELD;
+		}
 	}
 	return KR_STATE_DONE;
 }
 
+/** Change ranks of RRs from this single iteration: _INITIAL or _MISSING -> rank_to_set. */
+static void rank_records(kr_layer_t *ctx, enum kr_rank rank_to_set)
+{
+	struct kr_request *req	   = ctx->req;
+	struct kr_query *qry	   = req->current_query;
+	ranked_rr_array_t *ptrs[2] = { &req->answ_selected, &req->auth_selected };
+	for (size_t i = 0; i < 2; ++i) {
+		ranked_rr_array_t *arr = ptrs[i];
+		for (size_t j = 0; j < arr->len; ++j) {
+			ranked_rr_array_entry_t *entry = arr->at[j];
+			if (entry->qry_uid != qry->uid) {
+				continue;
+			}
+			if (kr_rank_test(entry->rank, KR_RANK_INITIAL)
+			    || kr_rank_test(entry->rank, KR_RANK_MISSING)) {
+				kr_rank_set(&entry->rank, rank_to_set);
+			}
+		}
+	}
+}
+
+static void check_wildcard(kr_layer_t *ctx)
+{
+	struct kr_request *req	   = ctx->req;
+	struct kr_query *qry	   = req->current_query;
+	ranked_rr_array_t *ptrs[2] = { &req->answ_selected, &req->auth_selected };
+
+	for (int i = 0; i < 2; ++i) {
+		ranked_rr_array_t *arr = ptrs[i];
+		for (ssize_t j = 0; j < arr->len; ++j) {
+			ranked_rr_array_entry_t *entry = arr->at[j];
+			const knot_rrset_t *rrsigs = entry->rr;
+
+			if (qry->uid != entry->qry_uid) {
+				continue;
+			}
+
+			if (rrsigs->type != KNOT_RRTYPE_RRSIG) {
+				continue;
+			}
+
+			int owner_labels = knot_dname_labels(rrsigs->owner, NULL);
+
+			for (int k = 0; k < rrsigs->rrs.rr_count; ++k) {
+				if (knot_rrsig_labels(&rrsigs->rrs, k) != owner_labels) {
+					qry->flags |= QUERY_DNSSEC_WEXPAND;
+				}
+			}
+		}
+	}
+}
+
 static int validate(kr_layer_t *ctx, knot_pkt_t *pkt)
 {
 	int ret = 0;
 	struct kr_request *req = ctx->req;
 	struct kr_query *qry = req->current_query;
+
 	/* Ignore faulty or unprocessed responses. */
 	if (ctx->state & (KR_STATE_FAIL|KR_STATE_CONSUME)) {
 		return ctx->state;
@@ -570,11 +794,30 @@ static int validate(kr_layer_t *ctx, knot_pkt_t *pkt)
 	/* Pass-through if user doesn't want secure answer or stub. */
 	/* @todo: Validating stub resolver mode. */
 	if (qry->flags & QUERY_STUB) {
+		rank_records(ctx, KR_RANK_OMIT);
 		return ctx->state;
 	}
+	uint8_t pkt_rcode = knot_wire_get_rcode(pkt->wire);
+	if ((qry->flags & QUERY_FORWARD) &&
+	    pkt_rcode != KNOT_RCODE_NOERROR &&
+	    pkt_rcode != KNOT_RCODE_NXDOMAIN) {
+		do {
+			qry->flags |= QUERY_DNSSEC_BOGUS;
+			if (qry->cname_parent) {
+				qry->cname_parent->flags |= QUERY_DNSSEC_BOGUS;
+			}
+			qry = qry->parent;
+		} while (qry);
+		ctx->state = KR_STATE_DONE;
+		return ctx->state;
+	}
+
 	if (!(qry->flags & QUERY_DNSSEC_WANT)) {
 		const uint32_t test_flags = (QUERY_CACHED | QUERY_DNSSEC_INSECURE);
 		const bool is_insec = ((qry->flags & test_flags) == test_flags);
+		if ((qry->flags & QUERY_DNSSEC_INSECURE)) {
+			rank_records(ctx, KR_RANK_INSECURE);
+		}
 		if (is_insec && qry->parent != NULL) {
 			/* We have got insecure answer from cache.
 			 * Mark parent(s) as insecure. */
@@ -593,6 +836,14 @@ static int validate(kr_layer_t *ctx, knot_pkt_t *pkt)
 
 	/* Pass-through if CD bit is set. */
 	if (knot_wire_get_cd(req->answer->wire)) {
+		check_wildcard(ctx);
+		/* Check if wildcard expansion happens.
+		 * If yes, copy authority. */
+		if ((qry->parent == NULL) &&
+		    (qry->flags & QUERY_DNSSEC_WEXPAND)) {
+			kr_ranked_rrarray_set_wire(&req->auth_selected, true, qry->uid, true);
+		}
+		rank_records(ctx, KR_RANK_OMIT);
 		return ctx->state;
 	}
 	/* Answer for RRSIG may not set DO=1, but all records MUST still validate. */
@@ -604,14 +855,19 @@ static int validate(kr_layer_t *ctx, knot_pkt_t *pkt)
 	}
 
 	/* Check if this is a DNSKEY answer, check trust chain and store. */
-	uint8_t pkt_rcode = knot_wire_get_rcode(pkt->wire);
 	uint16_t qtype = knot_pkt_qtype(pkt);
 	bool has_nsec3 = pkt_has_type(pkt, KNOT_RRTYPE_NSEC3);
 	const knot_pktsection_t *an = knot_pkt_section(pkt, KNOT_ANSWER);
 	const bool referral = (an->count == 0 && !knot_wire_get_aa(pkt->wire));
-	const bool no_data = (an->count == 0 && knot_wire_get_aa(pkt->wire));
 
 	if (!(qry->flags & QUERY_CACHED) && knot_wire_get_aa(pkt->wire)) {
+		/* Check if answer if not empty,
+		 * but iterator has not selected any records. */
+		if (!check_empty_answer(ctx, pkt)) {
+			VERBOSE_MSG(qry, "<= no useful RR in authoritative answer\n");
+			qry->flags |= QUERY_DNSSEC_BOGUS;
+			return KR_STATE_FAIL;
+		}
 		/* Track difference between current TA and signer name.
 		 * This indicates that the NS is auth for both parent-child,
 		 * and we must update DS/DNSKEY to validate it.
@@ -620,6 +876,10 @@ static int validate(kr_layer_t *ctx, knot_pkt_t *pkt)
 		if (ret != KR_STATE_DONE) {
 			return ret;
 		}
+		if ((qry->flags & (QUERY_FORWARD | QUERY_DNSSEC_INSECURE)) ==
+			(QUERY_FORWARD | QUERY_DNSSEC_INSECURE)) {
+			return KR_STATE_DONE;
+		}
 	}
 
 	if (knot_wire_get_aa(pkt->wire) && qtype == KNOT_RRTYPE_DNSKEY) {
@@ -627,6 +887,14 @@ static int validate(kr_layer_t *ctx, knot_pkt_t *pkt)
 		if (ret == kr_error(EAGAIN)) {
 			VERBOSE_MSG(qry, ">< cut changed, needs revalidation\n");
 			return KR_STATE_YIELD;
+		} else if (ret == kr_error(DNSSEC_INVALID_DS_ALGORITHM)) {
+			VERBOSE_MSG(qry, ">< all DS entries use unsupported algorithm pairs, going insecure\n");
+			/* ^ the message is a bit imprecise to avoid being too verbose */
+			qry->flags &= ~QUERY_DNSSEC_WANT;
+			qry->flags |= QUERY_DNSSEC_INSECURE;
+			rank_records(ctx, KR_RANK_INSECURE);
+			mark_insecure_parents(qry);
+			return KR_STATE_DONE;
 		} else if (ret != 0) {
 			VERBOSE_MSG(qry, "<= bad keys, broken trust chain\n");
 			qry->flags |= QUERY_DNSSEC_BOGUS;
@@ -635,14 +903,22 @@ static int validate(kr_layer_t *ctx, knot_pkt_t *pkt)
 	}
 
 	/* Validate non-existence proof if not positive answer. */
-	if (!(qry->flags & QUERY_CACHED) && pkt_rcode == KNOT_RCODE_NXDOMAIN) {
+	if (!(qry->flags & QUERY_CACHED) && pkt_rcode == KNOT_RCODE_NXDOMAIN &&
+	    ((qry->flags & (QUERY_FORWARD | QUERY_CNAME)) != (QUERY_FORWARD | QUERY_CNAME))) {
 		/* @todo If knot_pkt_qname(pkt) is used instead of qry->sname then the tests crash. */
 		if (!has_nsec3) {
 			ret = kr_nsec_name_error_response_check(pkt, KNOT_AUTHORITY, qry->sname);
 		} else {
 			ret = kr_nsec3_name_error_response_check(pkt, KNOT_AUTHORITY, qry->sname);
 		}
-		if (ret != 0) {
+		if (has_nsec3 && (ret == kr_error(DNSSEC_OUT_OF_RANGE))) {
+			/* NXDOMAIN proof is OK,
+			 * but NSEC3 that covers next closer name
+			 * (or wildcard at next closer name) has opt-out flag.
+			 * RFC5155 9.2; AD flag can not be set */
+			qry->flags |= QUERY_DNSSEC_OPTOUT;
+			VERBOSE_MSG(qry, "<= can't prove NXDOMAIN due to optout, going insecure\n");
+		} else if (ret != 0) {
 			VERBOSE_MSG(qry, "<= bad NXDOMAIN proof\n");
 			qry->flags |= QUERY_DNSSEC_BOGUS;
 			return KR_STATE_FAIL;
@@ -652,8 +928,10 @@ static int validate(kr_layer_t *ctx, knot_pkt_t *pkt)
 	/* @todo WTH, this needs API that just tries to find a proof and the caller
 	 * doesn't have to worry about NSEC/NSEC3
 	 * @todo rework this */
-	if (!(qry->flags & QUERY_CACHED)) {
-		if (pkt_rcode == KNOT_RCODE_NOERROR && no_data) {
+	if (!(qry->flags & QUERY_CACHED) && (pkt_rcode == KNOT_RCODE_NOERROR) &&
+	    ((qry->flags & (QUERY_FORWARD | QUERY_CNAME)) != (QUERY_FORWARD | QUERY_CNAME))) {
+		bool no_data = (an->count == 0 && knot_wire_get_aa(pkt->wire));
+		if (no_data) {
 			/* @todo
 			 * ? quick mechanism to determine which check to preform first
 			 * ? merge the functionality together to share code/resources
@@ -666,8 +944,7 @@ static int validate(kr_layer_t *ctx, knot_pkt_t *pkt)
 			if (ret != 0) {
 				if (has_nsec3 && (ret == kr_error(DNSSEC_OUT_OF_RANGE))) {
 					VERBOSE_MSG(qry, "<= can't prove NODATA due to optout, going insecure\n");
-					qry->flags &= ~QUERY_DNSSEC_WANT;
-					qry->flags |= QUERY_DNSSEC_INSECURE;
+					qry->flags |= QUERY_DNSSEC_OPTOUT;
 					/* Could not return from here,
 					 * we must continue, validate NSEC\NSEC3 and
 					 * call update_parent_keys() to mark
@@ -710,7 +987,7 @@ static int validate(kr_layer_t *ctx, knot_pkt_t *pkt)
 	/* Check if wildcard expansion detected for final query.
 	 * If yes, copy authority. */
 	if ((qry->parent == NULL) && (qry->flags & QUERY_DNSSEC_WEXPAND)) {
-		kr_ranked_rrarray_set_wire(&req->auth_selected, true, qry->uid);
+		kr_ranked_rrarray_set_wire(&req->auth_selected, true, qry->uid, true);
 	}
 
 	/* Check and update current delegation point security status. */
@@ -729,16 +1006,23 @@ static int validate(kr_layer_t *ctx, knot_pkt_t *pkt)
 		return KR_STATE_FAIL;
 	} else if (pkt_rcode == KNOT_RCODE_NOERROR &&
 		   referral &&
-		   ((qry->flags & (QUERY_DNSSEC_WANT | QUERY_DNSSEC_INSECURE)) == QUERY_DNSSEC_INSECURE)) {
+		   (((qry->flags & (QUERY_DNSSEC_WANT | QUERY_DNSSEC_INSECURE)) == QUERY_DNSSEC_INSECURE) ||
+		   (qry->flags & QUERY_DNSSEC_NODS))) {
 		/* referral with proven DS non-existance */
 		qtype = KNOT_RRTYPE_DS;
 	}
 	/* Update parent query zone cut */
 	if (qry->parent) {
-		if (update_parent_keys(qry, qtype) != 0) {
+		if (update_parent_keys(req, qtype) != 0) {
 			return KR_STATE_FAIL;
 		}
 	}
+
+	if (qry->flags & QUERY_FORWARD && qry->parent) {
+		if (pkt_rcode == KNOT_RCODE_NXDOMAIN) {
+			qry->parent->forward_flags |= QUERY_NO_MINIMIZE;
+		}
+	}
 	VERBOSE_MSG(qry, "<= answer valid, OK\n");
 	return KR_STATE_DONE;
 }
