@@ -44,30 +44,6 @@ static const struct kr_module embedded_modules[] = {
  #define LIBEXT ".so"
 #endif
 
-/** Check ABI version, return error on mismatch. */
-#define ABI_CHECK(m, prefix, symname, required) do { \
-	module_api_cb *_api = NULL; \
-	*(void **) (&_api) = load_symbol((m)->lib, (prefix), (symname)); \
-	if (_api == NULL) { \
-		return kr_error(ENOENT); \
-	} \
-	if (_api() != (required)) { \
-		return kr_error(ENOTSUP); \
-	} \
-} while (0)
-
-/** Load ABI by symbol names. */
-#define ABI_LOAD(m, prefix, s_init, s_deinit, s_config, s_layer, s_prop) do { \
-	module_prop_cb *module_prop = NULL; \
-	*(void **) (&(m)->init)   = load_symbol((m)->lib, (prefix), (s_init)); \
-	*(void **) (&(m)->deinit) = load_symbol((m)->lib, (prefix), (s_deinit)); \
-	*(void **) (&(m)->config) = load_symbol((m)->lib, (prefix), (s_config)); \
-	*(void **) (&(m)->layer)  = load_symbol((m)->lib, (prefix), (s_layer)); \
-	*(void **) (&module_prop) = load_symbol((m)->lib, (prefix), (s_prop)); \
-	if (module_prop != NULL) { \
-		(m)->props = module_prop(); \
-	} \
-} while(0)
 
 /** Load prefixed symbol. */
 static void *load_symbol(void *lib, const char *prefix, const char *name)
@@ -78,13 +54,9 @@ static void *load_symbol(void *lib, const char *prefix, const char *name)
 
 static int load_library(struct kr_module *module, const char *name, const char *path)
 {
+	assert(module && name && path);
 	/* Absolute or relative path (then only library search path is used). */
-	auto_free char *lib_path = NULL;
-	if (path != NULL) {
-		lib_path = kr_strcatdup(4, path, "/", name, LIBEXT);
-	} else {
-		lib_path = kr_strcatdup(2, name, LIBEXT);
-	}
+	auto_free char *lib_path = kr_strcatdup(4, path, "/", name, LIBEXT);
 	if (lib_path == NULL) {
 		return kr_error(ENOMEM);
 	}
@@ -115,9 +87,27 @@ static int load_sym_c(struct kr_module *module, uint32_t api_required)
 		}
 	}
 	/* Load dynamic library module */
-	auto_free char *module_prefix = kr_strcatdup(2, module->name, "_");
-	ABI_CHECK(module, module_prefix, "api", api_required);
-	ABI_LOAD(module, module_prefix, "init", "deinit", "config", "layer", "props");
+	auto_free char *m_prefix = kr_strcatdup(2, module->name, "_");
+
+	/* Check ABI version, return error on mismatch. */
+	module_api_cb *api = load_symbol(module->lib, m_prefix, "api");
+	if (api == NULL) {
+		return kr_error(ENOENT);
+	}
+	if (api() != api_required) {
+		return kr_error(ENOTSUP);
+	}
+
+	/* Load ABI by symbol names. */
+	#define ML(symname) module->symname = \
+		load_symbol(module->lib, m_prefix, #symname)
+	ML(init);
+	ML(deinit);
+	ML(config);
+	ML(layer);
+	ML(props);
+	#undef ML
+
 	return kr_ok();
 }
 
@@ -136,15 +126,9 @@ int kr_module_load(struct kr_module *module, const char *name, const char *path)
 		return kr_error(ENOMEM);
 	}
 
-	/* Search for module library, use current namespace if not found. */
-	if (load_library(module, name, path) != 0) {
-		/* Expand HOME env variable, as the linker may not expand it. */
-		auto_free char *local_path = kr_strcatdup(2, getenv("HOME"), "/.local/lib/kdns_modules");
-		if (load_library(module, name, local_path) != 0) {
-			if (load_library(module, name, MODULEDIR) != 0) {
-				module->lib = RTLD_DEFAULT;
-			}
-		}
+	/* Search for module library. */
+	if (!path || load_library(module, name, path) != 0) {
+		module->lib = RTLD_DEFAULT;
 	}
 
 	/* Try to load module ABI. */
