@@ -14,58 +14,77 @@
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
+/** @file
+ * Module API definition and functions for (un)loading modules.
+ */
+
 #pragma once
 
 #include "lib/defines.h"
 #include "lib/utils.h"
 #include "lib/layer.h"
 
-/*
- * Forward decls
- */
 struct kr_module;
 struct kr_prop;
 
-/*
- * API definition.
- * @cond internal
- */
-typedef uint32_t (module_api_cb)(void);
-typedef int (module_init_cb)(struct kr_module *);
-typedef int (module_deinit_cb)(struct kr_module *);
-typedef int (module_config_cb)(struct kr_module *, const char *);
-typedef const kr_layer_api_t* (module_layer_cb)(struct kr_module *);
-typedef struct kr_prop *(module_prop_cb)(void);
-typedef char *(kr_prop_cb)(void *, struct kr_module *, const char *);
-#define KR_MODULE_API ((uint32_t) 0x20161108)
-/* @endcond */
 
 /**
- * Module property (named callable).
- * A module property has a free-form JSON output (and optional input).
+ * Export module API version (place this at the end of your module).
+ *
+ * @param module module name (f.e. hints)
  */
-struct kr_prop {
-    kr_prop_cb *cb;
-    const char *name;
-    const char *info;
-};
+#define KR_MODULE_EXPORT(module) \
+    KR_EXPORT uint32_t module ## _api() { return KR_MODULE_API; }
+#define KR_MODULE_API ((uint32_t) 0x20161108)
+
+typedef uint32_t (module_api_cb)(void);
+
 
 /**
  * Module representation.
+ *
+ * The five symbols (init, ...) may be defined by the module as name_init(), etc;
+ * all are optional and missing symbols are represented as NULLs;
  */
 struct kr_module {
-    char *name;               /**< Name. */
-    module_init_cb   *init;   /**< Constructor */
-    module_deinit_cb *deinit; /**< Destructor */
-    module_config_cb *config; /**< Configuration */
-    module_layer_cb  *layer;  /**< Layer getter */
-    struct kr_prop   *props;  /**< Properties */
-    void *lib;                /**< Shared library handle or RTLD_DEFAULT */
-    void *data;               /**< Custom data context. */
+	char *name;
+
+	/** Constructor.  Called after loading the module.  @return error code. */
+	int (*init)(struct kr_module *self);
+	/** Destructor.  Called before unloading the module.  @return error code. */
+	int (*deinit)(struct kr_module *self);
+	/** Configure with encoded JSON (NULL if missing).  @return error code. */
+	int (*config)(struct kr_module *self, const char *input);
+	/** Get a pointer to packet processing API specs.  See docs on that type. */
+	const kr_layer_api_t * (*layer)(struct kr_module *self);
+	/** Get a pointer to list of properties, terminated by { NULL, NULL, NULL }. */
+	const struct kr_prop * (*props)(void);
+
+	void *lib;      /**< Shared library handle or RTLD_DEFAULT */
+	void *data;     /**< Custom data context. */
 };
 
 /**
- * Load module instance into memory.
+ * Module property callback.  Input and output is passed via a JSON encoded in a string.
+ *
+ * @param env pointer to the lua engine, i.e. struct engine *env (TODO: explicit type)
+ * @param input parameter (NULL if missing/nil on lua level)
+ * @return a free-form JSON output (malloc-ated)
+ */
+typedef char *(kr_prop_cb)(void *env, struct kr_module *self, const char *input);
+
+/**
+ * Module property (named callable).
+ */
+struct kr_prop {
+	kr_prop_cb *cb;
+	const char *name;
+	const char *info;
+};
+
+
+/**
+ * Load a C module instance into memory.
  *
  * @param module module structure
  * @param name module name
@@ -83,10 +102,3 @@ int kr_module_load(struct kr_module *module, const char *name, const char *path)
 KR_EXPORT
 void kr_module_unload(struct kr_module *module);
 
-/**
- * Export module API version (place this at the end of your module).
- *
- * @param module module name (f.e. hints)
- */
-#define KR_MODULE_EXPORT(module) \
-    KR_EXPORT uint32_t module ## _api() { return KR_MODULE_API; }
