@@ -27,6 +27,7 @@
 #include "lib/rplan.h"
 #include "lib/layer/iterate.h"
 #include "lib/dnssec/ta.h"
+#include "lib/dnssec.h"
 #if defined(ENABLE_COOKIES)
 #include "lib/cookies/control.h"
 #include "lib/cookies/helper.h"
@@ -37,6 +38,23 @@
 
 #define VERBOSE_MSG(qry, fmt...) QRVERBOSE((qry), "resl",  fmt)
 
+bool kr_rank_check(uint8_t rank)
+{
+	switch (rank & ~KR_RANK_AUTH) {
+	case KR_RANK_INITIAL:
+	case KR_RANK_OMIT:
+	case KR_RANK_INDET:
+	case KR_RANK_BOGUS:
+	case KR_RANK_MISMATCH:
+	case KR_RANK_MISSING:
+	case KR_RANK_INSECURE:
+	case KR_RANK_SECURE:
+		return true;
+	default:
+		return false;
+	}
+}
+
 /** @internal Set @a yielded to all RRs with matching @a qry_uid. */
 static void set_yield(ranked_rr_array_t *array, const uint32_t qry_uid, const bool yielded)
 {
@@ -131,14 +149,14 @@ static void randomized_qname_case(knot_dname_t * restrict qname, uint32_t secret
 static int invalidate_ns(struct kr_rplan *rplan, struct kr_query *qry)
 {
 	if (qry->ns.addr[0].ip.sa_family != AF_UNSPEC) {
-		uint8_t *addr = kr_nsrep_inaddr(qry->ns.addr[0]);
-		size_t addr_len = kr_nsrep_inaddr_len(qry->ns.addr[0]);
+		const char *addr = kr_inaddr(&qry->ns.addr[0].ip);
+		size_t addr_len = kr_inaddr_len(&qry->ns.addr[0].ip);
 		/* @warning _NOT_ thread-safe */
 		static knot_rdata_t rdata_arr[RDATA_ARR_MAX];
-		knot_rdata_init(rdata_arr, addr_len, addr, 0);
+		knot_rdata_init(rdata_arr, addr_len, (const uint8_t *)addr, 0);
 		return kr_zonecut_del(&qry->zone_cut, qry->ns.name, rdata_arr);
 	} else {
-		return kr_zonecut_del(&qry->zone_cut, qry->ns.name, NULL);
+		return kr_zonecut_del_all(&qry->zone_cut, qry->ns.name);
 	}
 }
 
@@ -180,13 +198,13 @@ static void check_empty_nonterms(struct kr_query *qry, knot_pkt_t *pkt, struct k
 static int ns_fetch_cut(struct kr_query *qry, const knot_dname_t *requested_name,
 			struct kr_request *req, knot_pkt_t *pkt)
 {
-	map_t *trust_anchors = &req->ctx->trust_anchors;
-	map_t *negative_anchors = &req->ctx->negative_anchors;
-
 	/* It can occur that here parent query already have
 	 * provably insecured zonecut which not in the cache yet. */
+	const uint32_t insec_flags = QUERY_DNSSEC_INSECURE | QUERY_DNSSEC_NODS;
+	const uint32_t cut_flags = QUERY_AWAIT_IPV4 | QUERY_AWAIT_IPV6;
 	const bool is_insecured = ((qry->parent != NULL) &&
-				   (qry->parent->flags & QUERY_DNSSEC_INSECURE) != 0);
+				   (qry->parent->flags & cut_flags) == 0 &&
+				   (qry->parent->flags & insec_flags) != 0);
 
 	/* Want DNSSEC if it's possible to secure this name
 	 * (e.g. is covered by any TA) */
@@ -195,8 +213,7 @@ static int ns_fetch_cut(struct kr_query *qry, const knot_dname_t *requested_name
 		 * even if cut name is covered by TA. */
 		qry->flags &= ~QUERY_DNSSEC_WANT;
 		qry->flags |= QUERY_DNSSEC_INSECURE;
-	} else if (!kr_ta_covers(negative_anchors, qry->zone_cut.name) &&
-		   kr_ta_covers(trust_anchors, qry->zone_cut.name)) {
+	} else if (kr_ta_covers_qry(req->ctx, qry->zone_cut.name, KNOT_RRTYPE_NS)) {
 		qry->flags |= QUERY_DNSSEC_WANT;
 	} else {
 		qry->flags &= ~QUERY_DNSSEC_WANT;
@@ -236,8 +253,7 @@ static int ns_fetch_cut(struct kr_query *qry, const knot_dname_t *requested_name
 	/* Zonecut name can change, check it again
 	 * to prevent unnecessary DS & DNSKEY queries */
 	if (!(qry->flags & QUERY_DNSSEC_INSECURE) &&
-	    !kr_ta_covers(negative_anchors, cut_found.name) &&
-	    kr_ta_covers(trust_anchors, cut_found.name)) {
+	    kr_ta_covers_qry(req->ctx, cut_found.name, KNOT_RRTYPE_NS)) {
 		qry->flags |= QUERY_DNSSEC_WANT;
 	} else {
 		qry->flags &= ~QUERY_DNSSEC_WANT;
@@ -247,7 +263,7 @@ static int ns_fetch_cut(struct kr_query *qry, const knot_dname_t *requested_name
 	    (cut_found.key == NULL)) {
 		/* No DNSKEY was found for cached cut.
 		 * If no glue were fetched for this cut,
-		 * we have got circular dependance - must fetch A\AAAA
+		 * we have got circular dependancy - must fetch A\AAAA
 		 * from authoritative, but we have no key to verify it.
 		 * TODO - try to refetch cut only if no glue were fetched */
 		kr_zonecut_deinit(&cut_found);
@@ -289,10 +305,12 @@ static int ns_resolve_addr(struct kr_query *qry, struct kr_request *param)
 	 * Prefer IPv6 and continue with IPv4 if not available.
 	 */
 	uint16_t next_type = 0;
-	if (!(qry->flags & QUERY_AWAIT_IPV6)) {
+	if (!(qry->flags & QUERY_AWAIT_IPV6) &&
+	    !(ctx->options & QUERY_NO_IPV6)) {
 		next_type = KNOT_RRTYPE_AAAA;
 		qry->flags |= QUERY_AWAIT_IPV6;
-	} else if (!(qry->flags & QUERY_AWAIT_IPV4)) {
+	} else if (!(qry->flags & QUERY_AWAIT_IPV4) &&
+		   !(ctx->options & QUERY_NO_IPV4)) {
 		next_type = KNOT_RRTYPE_A;
 		qry->flags |= QUERY_AWAIT_IPV4;
 		/* Hmm, no useable IPv6 then. */
@@ -315,10 +333,28 @@ static int ns_resolve_addr(struct kr_query *qry, struct kr_request *param)
 		invalidate_ns(rplan, qry);
 		return kr_error(EHOSTUNREACH);
 	}
-	/* Push new query to the resolution plan */
-	struct kr_query *next = kr_rplan_push(rplan, qry, qry->ns.name, KNOT_CLASS_IN, next_type);
-	if (!next) {
-		return kr_error(ENOMEM);
+	struct kr_query *next = qry;
+	if (knot_dname_is_equal(qry->ns.name, qry->sname) &&
+	    qry->stype == next_type) {
+		if (!(qry->flags & QUERY_NO_MINIMIZE)) {
+			qry->flags |= QUERY_NO_MINIMIZE;
+			qry->flags &= ~QUERY_AWAIT_IPV6;
+			qry->flags &= ~QUERY_AWAIT_IPV4;
+			VERBOSE_MSG(qry, "=> circular dependepcy, retrying with non-minimized name\n");
+		} else {
+			qry->ns.reputation |= KR_NS_NOIP4 | KR_NS_NOIP6;
+			kr_nsrep_update_rep(&qry->ns, qry->ns.reputation, ctx->cache_rep);
+			invalidate_ns(rplan, qry);
+			VERBOSE_MSG(qry, "=> unresolvable NS address, bailing out\n");
+			return kr_error(EHOSTUNREACH);
+		}
+	} else {
+		/* Push new query to the resolution plan */
+		next = kr_rplan_push(rplan, qry, qry->ns.name, KNOT_CLASS_IN, next_type);
+		if (!next) {
+			return kr_error(ENOMEM);
+		}
+		next->flags |= QUERY_NONAUTH;
 	}
 	/* At the root level with no NS addresses, add SBELT subrequest. */
 	int ret = 0;
@@ -388,8 +424,14 @@ static int edns_create(knot_pkt_t *pkt, knot_pkt_t *template, struct kr_request
 		wire_size += KR_COOKIE_OPT_MAX_LEN;
 	}
 #endif /* defined(ENABLE_COOKIES) */
-	if (req->has_tls && req->ctx->tls_padding >= 2) {
-		wire_size += KNOT_EDNS_OPTION_HDRLEN + req->ctx->tls_padding;
+	if (req->has_tls) {
+		if (req->ctx->tls_padding == -1)
+			/* FIXME: we do not know how to reserve space for the
+			 * default padding policy, since we can't predict what
+			 * it will select. So i'm just guessing :/ */
+			wire_size += KNOT_EDNS_OPTION_HDRLEN + 512;
+		if (req->ctx->tls_padding >= 2)
+			wire_size += KNOT_EDNS_OPTION_HDRLEN + req->ctx->tls_padding;
 	}
 	return knot_pkt_reserve(pkt, wire_size);
 }
@@ -413,28 +455,64 @@ static int answer_prepare(knot_pkt_t *answer, knot_pkt_t *query, struct kr_reque
 	return kr_ok();
 }
 
-static void write_extra_records(rr_array_t *arr, knot_pkt_t *answer)
+/** @return error code, ignoring if forced to truncate the packet. */
+static int write_extra_records(const rr_array_t *arr, knot_pkt_t *answer)
 {
 	for (size_t i = 0; i < arr->len; ++i) {
-		knot_pkt_put(answer, 0, arr->at[i], 0);
+		int err = knot_pkt_put(answer, 0, arr->at[i], 0);
+		if (err != KNOT_EOK) {
+			return err == KNOT_ESPACE ? kr_ok() : kr_error(err);
+		}
 	}
+	return kr_ok();
 }
 
-static void write_extra_ranked_records(ranked_rr_array_t *arr, knot_pkt_t *answer)
+/**
+ * @param all_secure optionally &&-combine security of written RRs into its value.
+ *		     (i.e. if you pass a pointer to false, it will always remain)
+ * @param all_cname optionally output if all written RRs are CNAMEs and RRSIGs of CNAMEs
+ * @return error code, ignoring if forced to truncate the packet.
+ */
+static int write_extra_ranked_records(const ranked_rr_array_t *arr, knot_pkt_t *answer,
+				      bool *all_secure, bool *all_cname)
 {
+	const bool has_dnssec = knot_pkt_has_dnssec(answer);
+	bool all_sec = true;
+	bool all_cn = (all_cname != NULL); /* optim.: init as false if not needed */
+	int err = kr_ok();
+
 	for (size_t i = 0; i < arr->len; ++i) {
 		ranked_rr_array_entry_t * entry = arr->at[i];
 		if (!entry->to_wire) {
 			continue;
 		}
 		knot_rrset_t *rr = entry->rr;
-		if (!knot_pkt_has_dnssec(answer)) {
+		if (!has_dnssec) {
 			if (rr->type != knot_pkt_qtype(answer) && knot_rrtype_is_dnssec(rr->type)) {
 				continue;
 			}
 		}
-		knot_pkt_put(answer, 0, rr, 0);
+		int err = knot_pkt_put(answer, 0, rr, 0);
+		if (err != KNOT_EOK) {
+			if (err == KNOT_ESPACE) {
+				err = kr_ok();
+			}
+			break;
+		}
+
+		if (rr->type != KNOT_RRTYPE_RRSIG) {
+			all_sec = all_sec && kr_rank_test(entry->rank, KR_RANK_SECURE);
+		}
+		all_cn = all_cn && kr_rrset_type_maysig(entry->rr) == KNOT_RRTYPE_CNAME;
 	}
+
+	if (all_secure) {
+		*all_secure = *all_secure && all_sec;
+	}
+	if (all_cname) {
+		*all_cname = all_cn;
+	}
+	return err;
 }
 
 /** @internal Add an EDNS padding RR into the answer if requested and required. */
@@ -444,17 +522,24 @@ static int answer_padding(struct kr_request *request)
 		assert(false);
 		return kr_error(EINVAL);
 	}
-	uint16_t padding = request->ctx->tls_padding;
+	int32_t padding = request->ctx->tls_padding;
 	knot_pkt_t *answer = request->answer;
 	knot_rrset_t *opt_rr = answer->opt_rr;
+	int32_t pad_bytes = -1;
 
-	if (padding < 2) {
-		return kr_ok();
+	if (padding == -1) { /* use the default padding policy from libknot */
+#if KNOT_VERSION_HEX < ((2 << 16) | (4 << 8) | 3)
+		/* no knot_edns_default_padding_size available in libknot */
+		padding = KR_DEFAULT_TLS_PADDING;
+#else
+		pad_bytes =  knot_edns_default_padding_size(answer, opt_rr);
+#endif
 	}
-	int32_t max_pad_bytes = knot_edns_get_payload(opt_rr) - (answer->size + knot_rrset_size(opt_rr));
-
-	int32_t pad_bytes = MIN(knot_edns_alignment_size(answer->size, knot_rrset_size(opt_rr), padding),
+	if (padding >= 2) {
+		int32_t max_pad_bytes = knot_edns_get_payload(opt_rr) - (answer->size + knot_rrset_size(opt_rr));
+		pad_bytes = MIN(knot_edns_alignment_size(answer->size, knot_rrset_size(opt_rr), padding),
 				max_pad_bytes);
+	}
 
 	if (pad_bytes >= 0) {
 		uint8_t zeros[MAX(1, pad_bytes)];
@@ -498,23 +583,46 @@ static int answer_finalize(struct kr_request *request, int state)
 		}
 	}
 
+	struct kr_query *last = rplan->resolved.len > 0 ? array_tail(rplan->resolved) : NULL;
+		/* TODO  ^^^^ this is slightly fragile */
+
+	/* AD flag.  We can only change `secure` from true to false.
+	 * Be conservative.  Primary approach: check ranks of all RRs in wire.
+	 * Only "negative answers" need special handling. */
+	bool secure = (last != NULL); /* suspicious otherwise */
+	if (last && (last->flags & QUERY_STUB)) {
+		secure = false; /* don't trust forwarding for now */
+	}
+	if (last && (last->flags & QUERY_DNSSEC_OPTOUT)) {
+		secure = false; /* the last answer is insecure due to opt-out */
+	}
+
+	bool answ_all_cnames = false/*arbitrary*/;
 	if (request->answ_selected.len > 0) {
 		assert(answer->current <= KNOT_ANSWER);
 		/* Write answer records. */
 		if (answer->current < KNOT_ANSWER) {
 			knot_pkt_begin(answer, KNOT_ANSWER);
 		}
-		write_extra_ranked_records(&request->answ_selected, answer);
+		if (write_extra_ranked_records(&request->answ_selected, answer,
+						&secure, &answ_all_cnames))
+		{
+			return answer_fail(request);
+		}
 	}
 
 	/* Write authority records. */
 	if (answer->current < KNOT_AUTHORITY) {
 		knot_pkt_begin(answer, KNOT_AUTHORITY);
 	}
-	write_extra_ranked_records(&request->auth_selected, answer);
+	if (write_extra_ranked_records(&request->auth_selected, answer, &secure, NULL)) {
+		return answer_fail(request);
+	}
 	/* Write additional records. */
 	knot_pkt_begin(answer, KNOT_ADDITIONAL);
-	write_extra_records(&request->additional, answer);
+	if (write_extra_records(&request->additional, answer)) {
+		return answer_fail(request);
+	}
 	/* Write EDNS information */
 	int ret = 0;
 	if (answer->opt_rr) {
@@ -527,18 +635,31 @@ static int answer_finalize(struct kr_request *request, int state)
 		ret = edns_put(answer);
 	}
 
-	/* Set AD=1 if succeeded and requested secured answer. */
-	const bool has_ad = knot_wire_get_ad(answer->wire);
-	knot_wire_clear_ad(answer->wire);
-	if (state == KR_STATE_DONE && rplan->resolved.len > 0) {
-		struct kr_query *last = array_tail(rplan->resolved);
-		/* Do not set AD for RRSIG query, as we can't validate it. */
-		const bool secure = (last->flags & QUERY_DNSSEC_WANT) &&
-		                   !(last->flags & QUERY_DNSSEC_INSECURE);
-		if (!(last->flags & QUERY_STUB) /* Never set AD if forwarding. */
-		    && has_ad && secure
-		    && knot_pkt_qtype(answer) != KNOT_RRTYPE_RRSIG) {
-			knot_wire_set_ad(answer->wire);
+	/* AD: "negative answers" need more handling. */
+	if (last && secure) {
+		if (kr_response_classify(answer) != PKT_NOERROR
+		    /* Additionally check for CNAME chains that "end in NODATA",
+		     * as those would also be PKT_NOERROR. */
+		    || (answ_all_cnames && knot_pkt_qtype(answer) != KNOT_RRTYPE_CNAME))
+		{
+			secure = secure && (last->flags & QUERY_DNSSEC_WANT)
+				&& !(last->flags & (QUERY_DNSSEC_BOGUS | QUERY_DNSSEC_INSECURE));
+		}
+	}
+	/* Clear AD if not secure.  ATM answer has AD=1 if requested secured answer. */
+	if (!secure || state != KR_STATE_DONE
+	    || knot_pkt_qtype(answer) == KNOT_RRTYPE_RRSIG) {
+		knot_wire_clear_ad(answer->wire);
+	}
+
+	if (last) {
+		struct kr_query *cname_parent = last->cname_parent;
+		while (cname_parent != NULL) {
+			if (cname_parent->flags & QUERY_DNSSEC_OPTOUT) {
+				knot_wire_clear_ad(answer->wire);
+				break;
+			}
+			cname_parent = cname_parent->cname_parent;
 		}
 	}
 
@@ -559,9 +680,17 @@ static int query_finalize(struct kr_request *request, struct kr_query *qry, knot
 			/* Stub resolution (ask for +rd and +do) */
 			if (qry->flags & QUERY_STUB) {
 				knot_wire_set_rd(pkt->wire);
-				if (knot_pkt_has_dnssec(request->answer))
+				if (knot_pkt_has_dnssec(request->answer)) {
 					knot_edns_set_do(pkt->opt_rr);
+				}
+				if (knot_wire_get_cd(request->answer->wire)) {
+					knot_wire_set_cd(pkt->wire);
+				}
 			/* Full resolution (ask for +cd and +do) */
+			} else if (qry->flags & QUERY_FORWARD) {
+				knot_wire_set_rd(pkt->wire);
+				knot_edns_set_do(pkt->opt_rr);
+				knot_wire_set_cd(pkt->wire);
 			} else if (qry->flags & QUERY_DNSSEC_WANT) {
 				knot_edns_set_do(pkt->opt_rr);
 				knot_wire_set_cd(pkt->wire);
@@ -615,10 +744,8 @@ static int resolve_query(struct kr_request *request, const knot_pkt_t *packet)
 	/* Deferred zone cut lookup for this query. */
 	qry->flags |= QUERY_AWAIT_CUT;
 	/* Want DNSSEC if it's posible to secure this name (e.g. is covered by any TA) */
-	map_t *negative_anchors = &request->ctx->negative_anchors;
-	map_t *trust_anchors = &request->ctx->trust_anchors;
 	if ((knot_wire_get_ad(packet->wire) || knot_pkt_has_dnssec(packet)) &&
-	    kr_ta_covers(trust_anchors, qname) && !kr_ta_covers(negative_anchors, qname)) {
+	    kr_ta_covers_qry(request->ctx, qname, qtype)) {
 		qry->flags |= QUERY_DNSSEC_WANT;
 	}
 
@@ -823,6 +950,201 @@ static struct kr_query *zone_cut_subreq(struct kr_rplan *rplan, struct kr_query
 	return next;
 }
 
+static int forward_trust_chain_check(struct kr_request *request, struct kr_query *qry, bool resume)
+{
+	struct kr_rplan *rplan = &request->rplan;
+	map_t *trust_anchors = &request->ctx->trust_anchors;
+	map_t *negative_anchors = &request->ctx->negative_anchors;
+
+	if (qry->parent != NULL &&
+	    !(qry->forward_flags & QUERY_CNAME) &&
+	    !(qry->flags & QUERY_DNS64_MARK) &&
+	    knot_dname_in(qry->parent->zone_cut.name, qry->zone_cut.name)) {
+		return KR_STATE_PRODUCE;
+	}
+
+	assert(qry->flags & QUERY_FORWARD);
+
+	if (!trust_anchors) {
+		qry->flags &= ~QUERY_AWAIT_CUT;
+		return KR_STATE_PRODUCE;
+	}
+
+	if (qry->flags & QUERY_DNSSEC_INSECURE) {
+		qry->flags &= ~QUERY_AWAIT_CUT;
+		return KR_STATE_PRODUCE;
+	}
+
+	if (qry->forward_flags & QUERY_NO_MINIMIZE) {
+		qry->flags &= ~QUERY_AWAIT_CUT;
+		return KR_STATE_PRODUCE;
+	}
+
+	const knot_dname_t *wanted_name = qry->sname;
+	const knot_dname_t *start_name = qry->sname;
+	if ((qry->flags & QUERY_AWAIT_CUT) && !resume) {
+		qry->flags &= ~QUERY_AWAIT_CUT;
+		const knot_dname_t *longest_ta = kr_ta_get_longest_name(trust_anchors, qry->sname);
+		if (longest_ta) {
+			start_name = longest_ta;
+			qry->zone_cut.name = knot_dname_copy(start_name, qry->zone_cut.pool);
+			qry->flags |= QUERY_DNSSEC_WANT;
+		} else {
+			qry->flags &= ~QUERY_DNSSEC_WANT;
+			return KR_STATE_PRODUCE;
+		}
+	}
+
+	bool has_ta = (qry->zone_cut.trust_anchor != NULL);
+	knot_dname_t *ta_name = (has_ta ? qry->zone_cut.trust_anchor->owner : NULL);
+	bool refetch_ta = (!has_ta || !knot_dname_is_equal(qry->zone_cut.name, ta_name));
+	bool is_dnskey_subreq = kr_rplan_satisfies(qry, ta_name, KNOT_CLASS_IN, KNOT_RRTYPE_DNSKEY);
+	bool refetch_key = has_ta && (!qry->zone_cut.key || !knot_dname_is_equal(ta_name, qry->zone_cut.key->owner));
+	if (refetch_key && !is_dnskey_subreq) {
+		struct kr_query *next = zone_cut_subreq(rplan, qry, ta_name, KNOT_RRTYPE_DNSKEY);
+		if (!next) {
+			return KR_STATE_FAIL;
+		}
+		return KR_STATE_DONE;
+	}
+
+	bool nods = false;
+	bool ds_req = false;
+	bool ns_req = false;
+	bool ns_exist = true;
+	bool minimized = false;
+	int name_offset = 1;
+	do {
+		wanted_name = start_name;
+		nods = false;
+		ds_req = false;
+		ns_req = false;
+		minimized = false;
+		ns_exist = true;
+
+		int cut_labels = knot_dname_labels(qry->zone_cut.name, NULL);
+		int wanted_name_labels = knot_dname_labels(wanted_name, NULL);
+		while (wanted_name[0] && wanted_name_labels > cut_labels + name_offset) {
+			wanted_name = knot_wire_next_label(wanted_name, NULL);
+			wanted_name_labels -= 1;
+		}
+		minimized = (wanted_name != qry->sname);
+
+		for (int i = 0; i < request->rplan.resolved.len; ++i) {
+			struct kr_query *q = request->rplan.resolved.at[i];
+			if (q->parent == qry &&
+			    q->sclass == qry->sclass &&
+			    (q->stype == KNOT_RRTYPE_DS || q->stype == KNOT_RRTYPE_NS) &&
+			    knot_dname_is_equal(q->sname, wanted_name)) {
+				if (q->stype == KNOT_RRTYPE_DS) {
+					ds_req = true;
+					if (q->flags & QUERY_DNSSEC_NODS) {
+						nods = true;
+					}
+					if (q->flags & QUERY_CNAME) {
+						nods = true;
+						ns_exist = false;
+					} else if (!(q->flags & QUERY_DNSSEC_OPTOUT)) {
+						int ret = kr_dnssec_matches_name_and_type(&request->auth_selected, q->uid,
+											  wanted_name, KNOT_RRTYPE_NS);
+						ns_exist = (ret == kr_ok());
+					}
+				} else {
+					if (q->flags & QUERY_CNAME) {
+						nods = true;
+						ns_exist = false;
+					}
+					ns_req = true;
+				}
+			}
+		}
+
+		if (ds_req && ns_exist && !ns_req && (minimized || resume)) {
+			struct kr_query *next = zone_cut_subreq(rplan, qry, wanted_name,
+								KNOT_RRTYPE_NS);
+			if (!next) {
+				return KR_STATE_FAIL;
+			}
+			return KR_STATE_DONE;
+		}
+
+		if (qry->parent == NULL && (qry->flags & QUERY_CNAME) &&
+		    ds_req && ns_req) {
+			return KR_STATE_PRODUCE;
+		}
+
+		if ((qry->stype == KNOT_RRTYPE_DS) &&
+	            knot_dname_is_equal(wanted_name, qry->sname)) {
+			nods = true;
+		} else if (resume && !ds_req) {
+			nods = false;
+		} else if (!minimized && qry->stype != KNOT_RRTYPE_DNSKEY) {
+			nods = true;
+		} else {
+			nods = ds_req;
+		}
+		name_offset += 1;
+	} while (ds_req && (ns_req || !ns_exist) && minimized);
+
+	/* Disable DNSSEC if it enters NTA. */
+	if (kr_ta_get(negative_anchors, wanted_name)){
+		VERBOSE_MSG(qry, ">< negative TA, going insecure\n");
+		qry->flags &= ~QUERY_DNSSEC_WANT;
+	}
+
+	/* Enable DNSSEC if enters a new island of trust. */
+	bool want_secured = (qry->flags & QUERY_DNSSEC_WANT) &&
+			    !knot_wire_get_cd(request->answer->wire);
+	if (!(qry->flags & QUERY_DNSSEC_WANT) &&
+	    !knot_wire_get_cd(request->answer->wire) &&
+	    kr_ta_get(trust_anchors, wanted_name)) {
+		qry->flags |= QUERY_DNSSEC_WANT;
+		want_secured = true;
+		WITH_VERBOSE {
+		char qname_str[KNOT_DNAME_MAXLEN];
+		knot_dname_to_str(qname_str, wanted_name, sizeof(qname_str));
+		VERBOSE_MSG(qry, ">< TA: '%s'\n", qname_str);
+		}
+	}
+
+	if (want_secured && !qry->zone_cut.trust_anchor) {
+		knot_rrset_t *ta_rr = kr_ta_get(trust_anchors, wanted_name);
+		if (!ta_rr) {
+			char name[] = "\0";
+			ta_rr = kr_ta_get(trust_anchors, (knot_dname_t*)name);
+		}
+		if (ta_rr) {
+			qry->zone_cut.trust_anchor = knot_rrset_copy(ta_rr, qry->zone_cut.pool);
+		}
+	}
+
+	has_ta = (qry->zone_cut.trust_anchor != NULL);
+	ta_name = (has_ta ? qry->zone_cut.trust_anchor->owner : NULL);
+	refetch_ta = (!has_ta || !knot_dname_is_equal(wanted_name, ta_name));
+	if (!nods && want_secured && refetch_ta) {
+		struct kr_query *next = zone_cut_subreq(rplan, qry, wanted_name,
+							KNOT_RRTYPE_DS);
+		if (!next) {
+			return KR_STATE_FAIL;
+		}
+		return KR_STATE_DONE;
+	}
+
+	/* Try to fetch missing DNSKEY.
+	 * Do not fetch if this is a DNSKEY subrequest to avoid circular dependency. */
+	is_dnskey_subreq = kr_rplan_satisfies(qry, ta_name, KNOT_CLASS_IN, KNOT_RRTYPE_DNSKEY);
+	refetch_key = has_ta && (!qry->zone_cut.key || !knot_dname_is_equal(ta_name, qry->zone_cut.key->owner));
+	if (want_secured && refetch_key && !is_dnskey_subreq) {
+		struct kr_query *next = zone_cut_subreq(rplan, qry, ta_name, KNOT_RRTYPE_DNSKEY);
+		if (!next) {
+			return KR_STATE_FAIL;
+		}
+		return KR_STATE_DONE;
+	}
+
+	return KR_STATE_PRODUCE;
+}
+
 /* @todo: Validator refactoring, keep this in driver for now. */
 static int trust_chain_check(struct kr_request *request, struct kr_query *qry)
 {
@@ -834,25 +1156,37 @@ static int trust_chain_check(struct kr_request *request, struct kr_query *qry)
 	if (kr_ta_get(negative_anchors, qry->zone_cut.name)){
 		VERBOSE_MSG(qry, ">< negative TA, going insecure\n");
 		qry->flags &= ~QUERY_DNSSEC_WANT;
+		qry->flags |= QUERY_DNSSEC_INSECURE;
 	}
-	/* Enable DNSSEC if enters a new island of trust. */
+	if (qry->flags & QUERY_DNSSEC_NODS) {
+		/* This is the next query iteration with minimized qname.
+		 * At previous iteration DS non-existance has been proven */
+		qry->flags &= ~QUERY_DNSSEC_NODS;
+		qry->flags &= ~QUERY_DNSSEC_WANT;
+		qry->flags |= QUERY_DNSSEC_INSECURE;
+	}
+	/* Enable DNSSEC if entering a new (or different) island of trust,
+	 * and update the TA RRset if required. */
 	bool want_secured = (qry->flags & QUERY_DNSSEC_WANT) &&
 			    !knot_wire_get_cd(request->answer->wire);
-	if (!(qry->flags & QUERY_DNSSEC_WANT) &&
-	    !knot_wire_get_cd(request->answer->wire) &&
-	    kr_ta_get(trust_anchors, qry->zone_cut.name)) {
+	knot_rrset_t *ta_rr = kr_ta_get(trust_anchors, qry->zone_cut.name);
+	if (!knot_wire_get_cd(request->answer->wire) && ta_rr) {
 		qry->flags |= QUERY_DNSSEC_WANT;
 		want_secured = true;
-		WITH_VERBOSE {
-		char qname_str[KNOT_DNAME_MAXLEN];
-		knot_dname_to_str(qname_str, qry->zone_cut.name, sizeof(qname_str));
-		VERBOSE_MSG(qry, ">< TA: '%s'\n", qname_str);
+
+		if (qry->zone_cut.trust_anchor == NULL
+		    || !knot_dname_is_equal(qry->zone_cut.trust_anchor->owner, qry->zone_cut.name)) {
+			mm_free(qry->zone_cut.pool, qry->zone_cut.trust_anchor);
+			qry->zone_cut.trust_anchor = knot_rrset_copy(ta_rr, qry->zone_cut.pool);
+
+			WITH_VERBOSE {
+			char qname_str[KNOT_DNAME_MAXLEN];
+			knot_dname_to_str(qname_str, ta_rr->owner, sizeof(qname_str));
+			VERBOSE_MSG(qry, ">< TA: '%s'\n", qname_str);
+			}
 		}
 	}
-	if (want_secured && !qry->zone_cut.trust_anchor) {
-		knot_rrset_t *ta_rr = kr_ta_get(trust_anchors, qry->zone_cut.name);
-		qry->zone_cut.trust_anchor = knot_rrset_copy(ta_rr, qry->zone_cut.pool);
-	}
+
 	/* Try to fetch missing DS (from above the cut). */
 	const bool has_ta = (qry->zone_cut.trust_anchor != NULL);
 	const knot_dname_t *ta_name = (has_ta ? qry->zone_cut.trust_anchor->owner : NULL);
@@ -863,7 +1197,7 @@ static int trust_chain_check(struct kr_request *request, struct kr_query *qry)
 		if (!next) {
 			return KR_STATE_FAIL;
 		}
-		next->flags |= QUERY_AWAIT_CUT|QUERY_DNSSEC_WANT;
+		next->flags |= (QUERY_AWAIT_CUT | QUERY_DNSSEC_WANT);
 		return KR_STATE_DONE;
 	}
 	/* Try to fetch missing DNSKEY (either missing or above current cut).
@@ -889,6 +1223,12 @@ static int zone_cut_check(struct kr_request *request, struct kr_query *qry, knot
 		return KR_STATE_PRODUCE;
 	}
 
+	/* Forwarding to upstream resolver mode.
+	 * Since forwarding targets already are in qry->ns -
+	 * cut fetching is not needed. */
+	if (qry->flags & QUERY_FORWARD) {
+		return forward_trust_chain_check(request, qry, false);
+	}
 	if (!(qry->flags & QUERY_AWAIT_CUT)) {
 		/* The query was resolved from cache.
 		 * Spawn DS \ DNSKEY requests if needed and exit */
@@ -957,7 +1297,15 @@ int kr_resolve_produce(struct kr_request *request, struct sockaddr **dst, int *t
 	struct kr_query *qry = array_tail(rplan->pending);
 	if (qry->deferred != NULL) {
 		/* @todo: Refactoring validator, check trust chain before resuming. */
-		switch(trust_chain_check(request, qry)) {
+		int state = 0;
+		if (((qry->flags & QUERY_FORWARD) == 0) ||
+		    ((qry->stype == KNOT_RRTYPE_DS) && (qry->flags & QUERY_CNAME))) {
+			state = trust_chain_check(request, qry);
+		} else {
+			state = forward_trust_chain_check(request, qry, true);
+		}
+
+		switch(state) {
 		case KR_STATE_FAIL: return KR_STATE_FAIL;
 		case KR_STATE_DONE: return KR_STATE_PRODUCE;
 		default: break;
@@ -968,7 +1316,10 @@ int kr_resolve_produce(struct kr_request *request, struct sockaddr **dst, int *t
 		set_yield(&request->answ_selected, qry->uid, false);
 		set_yield(&request->auth_selected, qry->uid, false);
 		RESUME_LAYERS(layer_id(request, pickle->api), request, qry, consume, pickle->pkt);
-		qry->deferred = pickle->next;
+		if (request->state != KR_STATE_YIELD) {
+			/* No new deferred answers, take the next */
+			qry->deferred = pickle->next;
+		}
 	} else {
 		/* Caller is interested in always tracking a zone cut, even if the answer is cached
 		 * this is normally not required, and incurrs another cache lookups for cached answer. */
@@ -983,7 +1334,7 @@ int kr_resolve_produce(struct kr_request *request, struct sockaddr **dst, int *t
 		request->state = KR_STATE_PRODUCE;
 		ITERATE_LAYERS(request, qry, produce, packet);
 		if (request->state != KR_STATE_FAIL && knot_wire_get_qr(packet->wire)) {
-			/* Produced an answer, consume it. */
+			/* Produced an answer from cache, consume it. */
 			qry->secret = 0;
 			request->state = KR_STATE_CONSUME;
 			ITERATE_LAYERS(request, qry, consume, packet);
@@ -1020,8 +1371,7 @@ int kr_resolve_produce(struct kr_request *request, struct sockaddr **dst, int *t
 
 ns_election:
 
-	/* If the query has got REFUSED & SERVFAIL, retry with current src up to KR_QUERY_NSRETRY_LIMIT.
-	 * If the query has already selected a NS and is waiting for IPv4/IPv6 record,
+	/* If the query has already selected a NS and is waiting for IPv4/IPv6 record,
 	 * elect best address only, otherwise elect a completely new NS.
 	 */
 	if(++ns_election_iter >= KR_ITER_LIMIT) {
@@ -1029,9 +1379,11 @@ ns_election:
 		return KR_STATE_FAIL;
 	}
 
-	const bool retry = (qry->flags & (QUERY_TCP|QUERY_STUB|QUERY_BADCOOKIE_AGAIN));
+	const bool retry = (qry->flags & (QUERY_TCP|QUERY_BADCOOKIE_AGAIN));
 	if (qry->flags & (QUERY_AWAIT_IPV4|QUERY_AWAIT_IPV6)) {
 		kr_nsrep_elect_addr(qry, request->ctx);
+	} else if (qry->flags & (QUERY_FORWARD|QUERY_STUB)) {
+		kr_nsrep_sort(&qry->ns, request->ctx->cache_rtt);
 	} else if (!qry->ns.name || !retry) { /* Keep NS when requerying/stub/badcookie. */
 		/* Root DNSKEY must be fetched from the hints to avoid chicken and egg problem. */
 		if (qry->sname[0] == '\0' && qry->stype == KNOT_RRTYPE_DNSKEY) {
@@ -1052,14 +1404,16 @@ ns_election:
 		int ret = ns_resolve_addr(qry, request);
 		if (ret != 0) {
 			qry->flags &= ~(QUERY_AWAIT_IPV6|QUERY_AWAIT_IPV4|QUERY_TCP);
+			qry->ns.name = NULL;
 			goto ns_election; /* Must try different NS */
 		}
 		ITERATE_LAYERS(request, qry, reset);
 		return KR_STATE_PRODUCE;
 	}
 
-	/* Randomize query case (if not in safemode) */
-	qry->secret = (qry->flags & QUERY_SAFEMODE) ? 0 : kr_rand_uint(UINT32_MAX);
+	/* Randomize query case (if not in safemode or turned off) */
+	qry->secret = (qry->flags & (QUERY_SAFEMODE | QUERY_NO_0X20))
+			? 0 : kr_rand_uint(0);
 	knot_dname_t *qname_raw = (knot_dname_t *)knot_pkt_qname(packet);
 	randomized_qname_case(qname_raw, qry->secret);
 
@@ -1155,7 +1509,7 @@ int kr_resolve_checkout(struct kr_request *request, struct sockaddr *src,
 		if (!kr_inaddr_equal(dst, addr)) {
 			continue;
 		}
-		inet_ntop(addr->sa_family, kr_nsrep_inaddr(qry->ns.addr[i]), ns_str, sizeof(ns_str));
+		inet_ntop(addr->sa_family, kr_inaddr(&qry->ns.addr[i].ip), ns_str, sizeof(ns_str));
 		VERBOSE_MSG(qry, "=> querying: '%s' score: %u zone cut: '%s' m12n: '%s' type: '%s' proto: '%s'\n",
 			ns_str, qry->ns.score, zonecut_str, qname_str, type_str, (qry->flags & QUERY_TCP) ? "tcp" : "udp");
 		break;
