@@ -34,7 +34,7 @@
 	X(AWAIT_IPV4,      1 << 6) /**< Query is waiting for A address. */ \
 	X(AWAIT_IPV6,      1 << 7) /**< Query is waiting for AAAA address. */ \
 	X(AWAIT_CUT,       1 << 8) /**< Query is waiting for zone cut lookup */ \
-	X(SAFEMODE,        1 << 9) /**< Don't use fancy stuff (EDNS...) */ \
+	X(SAFEMODE,        1 << 9) /**< Don't use fancy stuff (EDNS, 0x20, ...) */ \
 	X(CACHED,          1 << 10) /**< Query response is cached. */ \
 	X(NO_CACHE,        1 << 11) /**< No cache for lookup; exception: finding NSs and subqueries. */ \
 	X(EXPIRING,        1 << 12) /**< Query response is cached, but expiring. */ \
@@ -51,8 +51,14 @@
 	X(BADCOOKIE_AGAIN, 1 << 22) /**< Query again because bad cookie returned. */ \
 	X(CNAME,           1 << 23) /**< Query response contains CNAME in answer section. */ \
 	X(REORDER_RR,      1 << 24) /**< Reorder cached RRs. */ \
-	X(TRACE,	   1 << 25) /**< Log answer with kr_verbose_log(), unless -DNDEBUG. */
-			/* 1 << 31       Used by ../modules/dns64/dns64.lua */
+	X(TRACE,	   1 << 25) /**< Log answer with kr_verbose_log(), unless -DNDEBUG. */ \
+	X(NO_0X20,	   1 << 26) /**< Disable query case randomization . */ \
+	X(DNSSEC_NODS,	   1 << 27) /**< DS non-existance is proven */ \
+	X(DNSSEC_OPTOUT,   1 << 28) /**< Closest encloser proof has optout */ \
+	X(NONAUTH,         1 << 29) /**< Non-authoritative in-bailiwick records are enough.
+				     *   TODO: utilize this also outside cache. */ \
+	X(FORWARD,	   1 << 30) /**< Forward all queries to upstream; validate answers */ \
+	X(DNS64_MARK,	   1u << 31) /**< Internal to ../modules/dns64/dns64.lua */
 
 /** Query flags */
 enum kr_query_flag {
@@ -83,6 +89,9 @@ struct kr_query {
 	struct kr_nsrep ns;
 	struct kr_layer_pickle *deferred;
 	uint32_t uid; /**< Query iteration number, unique within the kr_rplan. */
+	uint32_t forward_flags;
+	/** Pointer to the query that originated this one because of following a CNAME (or NULL). */
+	struct kr_query *cname_parent;
 };
 
 /** @cond internal Array of queries. */
@@ -176,3 +185,16 @@ struct kr_query *kr_rplan_resolved(struct kr_rplan *rplan);
 /** Return query predecessor. */
 KR_EXPORT KR_PURE
 struct kr_query *kr_rplan_next(struct kr_query *qry);
+
+/**
+ * Check if a given query already resolved.
+ * @param rplan plan instance
+ * @param parent query parent (or NULL)
+ * @param name resolved name
+ * @param cls  resolved class
+ * @param type resolved type
+ * @return query instance or NULL
+ */
+KR_EXPORT KR_PURE
+struct kr_query *kr_rplan_find_resolved(struct kr_rplan *rplan, struct kr_query *parent,
+                               const knot_dname_t *name, uint16_t cls, uint16_t type);
