@@ -27,6 +27,7 @@
 #include <libknot/rrtype/rrsig.h>
 #include <libknot/rrset-dump.h>
 #include <libknot/version.h>
+#include <uv.h>
 
 #include "lib/defines.h"
 #include "lib/utils.h"
@@ -126,16 +127,25 @@ char* kr_strcatdup(unsigned n, ...)
 	return result;
 }
 
-static int seed_file(FILE *fp, char *buf, size_t buflen)
+static int seed_file(const char *fname, char *buf, size_t buflen)
 {
+	auto_fclose FILE *fp = fopen(fname, "r");
 	if (!fp) {
-		return -1;
-	}
-	/* Read whole buffer even if interrupted */
-	ssize_t readb = 0;
-	while (!ferror(fp) && readb < buflen) {
-		readb += fread(buf, 1, buflen - readb, fp);
+		return kr_error(EINVAL);
 	}
+	/* Disable buffering to conserve randomness but ignore failing to do so. */
+	setvbuf(fp, NULL, _IONBF, 0);
+	do {
+		if (feof(fp)) {
+			return kr_error(ENOENT);
+		}
+		if (ferror(fp)) {
+			return kr_error(ferror(fp));
+		}
+		if (fread(buf, buflen, 1, fp) == 1) { /* read in one chunk for simplicity */
+			return kr_ok();
+		}
+	} while (true);
 	return 0;
 }
 
@@ -146,13 +156,13 @@ static int randseed(char *buf, size_t buflen)
         "/dev/srandom", "/dev/urandom", "/dev/random", NULL
     };
     for (unsigned i = 0; filenames[i]; ++i) {
-        auto_fclose FILE *fp = fopen(filenames[i], "r");
-        if (seed_file(fp, buf, buflen) == 0) {
+        if (seed_file(filenames[i], buf, buflen) == 0) {
             return 0;
         }
     }
 
     /* Seed from time, this is not going to be secure. */
+    kr_log_error("failed to obtain randomness, falling back to current time\n");
     struct timeval tv;
     gettimeofday(&tv, NULL);
     memcpy(buf, &tv, buflen < sizeof(tv) ? buflen : sizeof(tv));
@@ -167,13 +177,15 @@ int kr_rand_reseed(void)
 	return kr_ok();
 }
 
-unsigned kr_rand_uint(unsigned max)
+uint32_t kr_rand_uint(uint32_t max)
 {
-	if (!isaac_seeded) {
+	if (unlikely(!isaac_seeded)) {
 		kr_rand_reseed();
 		isaac_seeded = true;
 	}
-	return isaac_next_uint(&ISAAC, max);
+	return max == 0
+		? isaac_next_uint32(&ISAAC)
+		: isaac_next_uint(&ISAAC, max);
 }
 
 int kr_memreserve(void *baton, char **mem, size_t elm_size, size_t want, size_t *have)
@@ -244,6 +256,13 @@ int kr_pkt_put(knot_pkt_t *pkt, const knot_dname_t *name, uint32_t ttl,
 	return knot_pkt_put(pkt, 0, &rr, KNOT_PF_FREE);
 }
 
+void kr_pkt_make_auth_header(knot_pkt_t *pkt)
+{
+	assert(pkt && pkt->wire);
+	knot_wire_clear_ad(pkt->wire);
+	knot_wire_set_aa(pkt->wire);
+}
+
 const char *kr_inaddr(const struct sockaddr *addr)
 {
 	if (!addr) {
@@ -271,6 +290,18 @@ int kr_inaddr_len(const struct sockaddr *addr)
 	return kr_family_len(addr->sa_family);
 }
 
+uint16_t kr_inaddr_port(const struct sockaddr *addr)
+{
+	if (!addr) {
+		return 0;
+	}
+	switch (addr->sa_family) {
+	case AF_INET:  return ntohs(((const struct sockaddr_in *)addr)->sin_port);
+	case AF_INET6: return ntohs(((const struct sockaddr_in6 *)addr)->sin6_port);
+	default:       return 0;
+	}
+}
+
 int kr_straddr_family(const char *addr)
 {
 	if (!addr) {
@@ -291,6 +322,32 @@ int kr_family_len(int family)
 	}
 }
 
+struct sockaddr * kr_straddr_socket(const char *addr, int port)
+{
+	switch (kr_straddr_family(addr)) {
+	case AF_INET: {
+		struct sockaddr_in *res = malloc(sizeof(*res));
+		if (uv_ip4_addr(addr, port, res) >= 0) {
+			return (struct sockaddr *)res;
+		} else {
+			free(res);
+			return NULL;
+		}
+	}
+	case AF_INET6: {
+		struct sockaddr_in6 *res = malloc(sizeof(*res));
+		if (uv_ip6_addr(addr, port, res) >= 0) {
+			return (struct sockaddr *)res;
+		} else {
+			free(res);
+			return NULL;
+		}
+	}
+	default:
+		return NULL;
+	}
+}
+
 int kr_straddr_subnet(void *dst, const char *addr)
 {
 	if (!dst || !addr) {
@@ -325,9 +382,19 @@ int kr_straddr_subnet(void *dst, const char *addr)
 
 int kr_bitcmp(const char *a, const char *b, int bits)
 {
-	if (!a || !b || bits == 0) {
-		return kr_error(ENOMEM);
+	/* We're using the function from lua directly, so at least for now
+	 * we avoid crashing on bogus inputs.  Meaning: NULL is ordered before
+	 * anything else, and negative length is the same as zero.
+	 * TODO: review the call sites and probably remove the checks. */
+	if (bits <= 0 || (!a && !b)) {
+		return 0;
+	} else if (!a) {
+		return -1;
+	} else if (!b) {
+		return 1;
 	}
+
+	assert((a && b && bits >= 0)  ||  bits == 0);
 	/* Compare part byte-divisible part. */
 	const size_t chunk = bits / 8;
 	int ret = memcmp(a, b, chunk);
@@ -373,11 +440,10 @@ int kr_rrmap_add(map_t *stash, const knot_rrset_t *rr, uint8_t rank, knot_mm_t *
 	/* Stash key = {[1] flags, [1-255] owner, [5] type, [1] \x00 } */
 	char key[KR_RRKEY_LEN];
 	uint8_t extra_flags = 0;
-	uint16_t rrtype = rr->type;
+	uint16_t rrtype = kr_rrset_type_maysig(rr);
 	/* Stash RRSIGs in a special cache, flag them and set type to its covering RR.
 	 * This way it the stash won't merge RRSIGs together. */
 	if (rr->type == KNOT_RRTYPE_RRSIG) {
-		rrtype = knot_rrsig_type_covered(&rr->rrs, 0);
 		extra_flags |= KEY_FLAG_RRSIG;
 	}
 	int ret = kr_rrkey(key, rr->owner, rrtype, rank);
@@ -399,17 +465,48 @@ int kr_rrmap_add(map_t *stash, const knot_rrset_t *rr, uint8_t rank, knot_mm_t *
 	return knot_rdataset_merge(&stashed->rrs, &rr->rrs, pool);
 }
 
-int kr_rrarray_add(rr_array_t *array, const knot_rrset_t *rr, knot_mm_t *pool)
+/** Return whether two RRsets match, i.e. would form the same set; see ranked_rr_array_t */
+static inline bool rrsets_match(const knot_rrset_t *rr1, const knot_rrset_t *rr2)
 {
-	int ret = array_reserve_mm(*array, array->len + 1, kr_memreserve, pool);
-	if (ret != 0) {
-		return kr_error(ENOMEM);
+	bool match = rr1->type == rr2->type && rr1->rclass == rr2->rclass;
+	if (match && rr2->type == KNOT_RRTYPE_RRSIG) {
+		match = match && knot_rrsig_type_covered(&rr1->rrs, 0)
+				  == knot_rrsig_type_covered(&rr2->rrs, 0);
 	}
-	knot_rrset_t *copy = knot_rrset_copy(rr, pool);
-	if (!copy) {
-		return kr_error(ENOMEM);
+	match = match && knot_dname_is_equal(rr1->owner, rr2->owner);
+	return match;
+}
+
+/** Ensure that an index in a ranked array won't cause "duplicate" RRsets on wire.
+ *
+ * Other entries that would form the same RRset get to_wire = false.
+ * See also rrsets_match.
+ */
+static int to_wire_ensure_unique(ranked_rr_array_t *array, size_t index)
+{
+	bool ok = array && index < array->len;
+	if (!ok) {
+		assert(false);
+		return kr_error(EINVAL);
+	}
+
+	const struct ranked_rr_array_entry *e0 = array->at[index];
+	if (!e0->to_wire) {
+		return kr_ok();
+	}
+
+	for (ssize_t i = array->len - 1; i >= 0; --i) {
+		/* ^ iterate backwards, as the end is more likely in CPU caches */
+		struct ranked_rr_array_entry *ei = array->at[i];
+		if (ei->qry_uid == e0->qry_uid /* assumption: no duplicates within qry */
+		    || !ei->to_wire /* no use for complex comparison if @to_wire */
+		   ) {
+			continue;
+		}
+		if (rrsets_match(ei->rr, e0->rr)) {
+			ei->to_wire = false;
+		}
 	}
-	array_push(*array, copy);
 	return kr_ok();
 }
 
@@ -428,15 +525,18 @@ int kr_ranked_rrarray_add(ranked_rr_array_t *array, const knot_rrset_t *rr,
 		if (stashed->qry_uid != qry_uid) {
 			break;
 		}
-		if (stashed->rr->rclass == rr->rclass &&
-		    stashed->rr->type == rr->type &&
-		    knot_dname_is_equal(stashed->rr->owner, rr->owner)) {
-			assert(stashed->rank == rank &&
-			       stashed->cached == false &&
-			       stashed->to_wire == to_wire);
-			/* Merge */
-			return knot_rdataset_merge(&stashed->rr->rrs, &rr->rrs, pool);
+		if (!rrsets_match(stashed->rr, rr)) {
+			continue;
 		}
+		/* Found the entry to merge with.  Check consistency and merge. */
+		bool ok = stashed->rank == rank
+			&& !stashed->cached
+			&& stashed->to_wire == to_wire;
+		if (!ok) {
+			assert(false);
+			return kr_error(EEXIST);
+		}
+		return knot_rdataset_merge(&stashed->rr->rrs, &rr->rrs, pool);
 	}
 
 	/* No stashed rrset found, add */
@@ -462,16 +562,28 @@ int kr_ranked_rrarray_add(ranked_rr_array_t *array, const knot_rrset_t *rr,
 	entry->cached = false;
 	entry->yielded = false;
 	entry->to_wire = to_wire;
-	array_push(*array, entry);
-	return kr_ok();
+	if (array_push(*array, entry) < 0) {
+		/* Silence coverity.  It shouldn't be possible to happen,
+		 * due to the array_reserve_mm call above. */
+		mm_free(pool, entry);
+		return kr_error(ENOMEM);
+	}
+
+	return to_wire_ensure_unique(array, array->len - 1);
 }
 
-int kr_ranked_rrarray_set_wire(ranked_rr_array_t *array, bool to_wire, uint32_t qry_uid)
+int kr_ranked_rrarray_set_wire(ranked_rr_array_t *array, bool to_wire,
+			       uint32_t qry_uid, bool check_dups)
 {
 	for (size_t i = 0; i < array->len; ++i) {
 		ranked_rr_array_entry_t *entry = array->at[i];
-		if (entry->qry_uid == qry_uid) {
-			entry->to_wire = to_wire;
+		if (entry->qry_uid != qry_uid) {
+			continue;
+		}
+		entry->to_wire = to_wire;
+		if (check_dups) {
+			int ret = to_wire_ensure_unique(array, i);
+			if (ret) return ret;
 		}
 	}
 	return kr_ok();
@@ -480,10 +592,10 @@ int kr_ranked_rrarray_set_wire(ranked_rr_array_t *array, bool to_wire, uint32_t
 
 static char *callprop(struct kr_module *module, const char *prop, const char *input, void *env)
 {
-	if (!module || !prop) {
+	if (!module || !module->props || !prop) {
 		return NULL;
 	}
-	for (struct kr_prop *p = module->props; p && p->name; ++p) {
+	for (const struct kr_prop *p = module->props(); p && p->name; ++p) {
 		if (p->cb != NULL && strcmp(p->name, prop) == 0) {
 			return p->cb(env, module, input);
 		}
@@ -534,8 +646,8 @@ static void flags_to_str(char *dst, const knot_pkt_t *pkt, size_t maxlen)
 		{knot_wire_get_rd, "rd"},
 		{knot_wire_get_ra, "ra"},
 		{knot_wire_get_tc, "tc"},
-		{knot_wire_get_cd, "cd"},
-		{knot_wire_get_ad, "ad"}
+		{knot_wire_get_ad, "ad"},
+		{knot_wire_get_cd, "cd"}
 	};
 	for (int i = 0; i < 7; ++i) {
 		if (!flag[i].get(pkt->wire)) {
@@ -547,7 +659,7 @@ static void flags_to_str(char *dst, const knot_pkt_t *pkt, size_t maxlen)
 			return;
 		}
 		offset += ret;
-		maxlen -= offset;
+		maxlen -= ret;
 	}
 	dst[offset] = 0;
 }
@@ -647,15 +759,24 @@ void kr_pkt_print(knot_pkt_t *pkt)
 
 void kr_dname_print(const knot_dname_t *name, const char *prefix, const char *postfix)
 {
-	char str[KNOT_DNAME_MAXLEN];
+	char str[KNOT_DNAME_MAXLEN] = {0};
 	knot_dname_to_str(str, name, KNOT_DNAME_MAXLEN);
-	kr_log_verbose ("%s%s%s", prefix, str, postfix);
+	kr_log_verbose("%s%s%s", prefix, str, postfix);
 }
 
 void kr_rrtype_print(const uint16_t rrtype, const char *prefix, const char *postfix)
 {
-	char str[32];
+	char str[32] = {0};
 	knot_rrtype_to_string(rrtype, str, 32);
-	kr_log_verbose ("%s%s%s", prefix, str, postfix);
+	kr_log_verbose("%s%s%s", prefix, str, postfix);
+}
+
+void kr_qry_print(const struct kr_query *qry, const char *prefix, const char *postfix)
+{
+	char str[6] = {0};
+	knot_rrclass_to_string(qry->sclass, str, sizeof(str));
+	kr_dname_print(qry->sname, prefix, " ");
+	kr_log_verbose("%s",str);
+	kr_rrtype_print(qry->stype, " ", postfix);
 }
 
