@@ -19,13 +19,19 @@
 #include <assert.h>
 #include <stdio.h>
 #include <stdbool.h>
+#include <sys/socket.h>
 #include <sys/time.h>
 #include <netinet/in.h>
+#include <libknot/libknot.h>
 #include <libknot/packet/pkt.h>
+#include <libknot/rrset.h>
+#include <libknot/rrtype/rrsig.h>
 #include "lib/generic/map.h"
 #include "lib/generic/array.h"
 #include "lib/defines.h"
 
+struct kr_query;
+
 /*
  * Logging and debugging.
  */
@@ -66,12 +72,30 @@ static inline void *mm_alloc(knot_mm_t *mm, size_t size)
 }
 static inline void mm_free(knot_mm_t *mm, void *what)
 {
-	if (mm) { 
+	if (mm) {
 		if (mm->free)
 			mm->free(what);
 	}
 	else free(what);
 }
+static inline void *mm_realloc(knot_mm_t *mm, void *what, size_t size, size_t prev_size)
+{
+	if (mm) {
+		void *p = mm->alloc(mm->ctx, size);
+		if (p == NULL) {
+			return NULL;
+		} else {
+			if (what) {
+				memcpy(p, what,
+				       prev_size < size ? prev_size : size);
+			}
+			mm_free(mm, what);
+			return p;
+		}
+	} else {
+		return realloc(what, size);
+	}
+}
 /* @endcond */
 
 /** Return time difference in miliseconds.
@@ -90,17 +114,26 @@ static inline long time_diff(struct timeval *begin, struct timeval *end) {
 
 /** @cond internal Array types */
 struct kr_context;
+
 typedef array_t(knot_rrset_t *) rr_array_t;
 struct ranked_rr_array_entry {
 	uint32_t qry_uid;
-	uint8_t rank;
+	uint8_t rank; /**< enum kr_rank */
 	uint8_t revalidation_cnt;
-	bool cached;
+	bool cached;  /**< whether it has been stashed to cache already */
 	bool yielded;
-	bool to_wire;
+	bool to_wire; /**< whether to be put into the answer */
 	knot_rrset_t *rr;
 };
 typedef struct ranked_rr_array_entry ranked_rr_array_entry_t;
+
+/** Array of RRsets coming from multiple queries; for struct kr_request.
+ *
+ * Notes:
+ *  - RRSIGs are only considered to form an RRset when the types covered match;
+ *    cache-related code relies on that!
+ *  - RRsets from the same packet (qry_uid) get merged.
+ */
 typedef array_t(ranked_rr_array_entry_t *) ranked_rr_array_t;
 /* @endcond */
 
@@ -116,9 +149,12 @@ char* kr_strcatdup(unsigned n, ...);
 /** Reseed CSPRNG context. */
 int kr_rand_reseed(void);
 
-/** Get pseudo-random value. */
+/** Get pseudo-random value between zero and max-1 (inclusive).
+ *
+ * Passing zero means that any uint32_t should be returned (it's also faster).
+ */
 KR_EXPORT
-unsigned kr_rand_uint(unsigned max);
+uint32_t kr_rand_uint(uint32_t max);
 
 /** Memory reservation routine for knot_mm_t */
 KR_EXPORT
@@ -137,6 +173,17 @@ KR_EXPORT
 int kr_pkt_put(knot_pkt_t *pkt, const knot_dname_t *name, uint32_t ttl,
                uint16_t rclass, uint16_t rtype, const uint8_t *rdata, uint16_t rdlen);
 
+/** Set packet header suitable for authoritative answer. (for policy module) */
+KR_EXPORT
+void kr_pkt_make_auth_header(knot_pkt_t *pkt);
+
+/** Simple storage for IPx address or AF_UNSPEC. */
+union inaddr {
+	struct sockaddr ip;
+	struct sockaddr_in ip4;
+	struct sockaddr_in6 ip6;
+};
+
 /** Address bytes for given family. */
 KR_EXPORT KR_PURE
 const char *kr_inaddr(const struct sockaddr *addr);
@@ -146,24 +193,35 @@ int kr_inaddr_family(const struct sockaddr *addr);
 /** Address length for given family. */
 KR_EXPORT KR_PURE
 int kr_inaddr_len(const struct sockaddr *addr);
+/** Port. */
+KR_EXPORT KR_PURE
+uint16_t kr_inaddr_port(const struct sockaddr *addr);
 /** Return address type for string. */
 KR_EXPORT KR_PURE
 int kr_straddr_family(const char *addr);
 /** Return address length in given family. */
 KR_EXPORT KR_CONST
 int kr_family_len(int family);
+/** Create a sockaddr* from string+port representation (also accepts IPv6 link-local). */
+KR_EXPORT
+struct sockaddr * kr_straddr_socket(const char *addr, int port);
 /** Parse address and return subnet length (bits).
   * @warning 'dst' must be at least `sizeof(struct in6_addr)` long. */
 KR_EXPORT
 int kr_straddr_subnet(void *dst, const char *addr);
-/** Compare memory bitwise. */
+
+/** Compare memory bitwise.  The semantics is "the same" as for memcmp().
+ *  The partial byte is considered with more-significant bits first,
+ *  so this is e.g. suitable for comparing IP prefixes. */
 KR_EXPORT KR_PURE
 int kr_bitcmp(const char *a, const char *b, int bits);
 
 /** @internal RR map flags. */
-#define KEY_FLAG_RRSIG 0x02
-#define KEY_FLAG_RANK(key) (key[0] >> 2)
-#define KEY_COVERING_RRSIG(key) (key[0] & KEY_FLAG_RRSIG)
+static const uint8_t KEY_FLAG_RRSIG = 0x02;
+static inline uint8_t KEY_FLAG_RANK(const char *key)
+	{ return ((uint8_t)(key[0])) >> 2; }
+static inline bool KEY_COVERING_RRSIG(const char *key)
+	{ return ((uint8_t)(key[0])) & KEY_FLAG_RRSIG; }
 
 /* Stash key = {[1] flags, [1-255] owner, [5] type, [1] \x00 } */
 #define KR_RRKEY_LEN (9 + KNOT_DNAME_MAXLEN)
@@ -183,16 +241,24 @@ int kr_rrkey(char *key, const knot_dname_t *owner, uint16_t type, uint8_t rank);
  */
 int kr_rrmap_add(map_t *stash, const knot_rrset_t *rr, uint8_t rank, knot_mm_t *pool);
 
-/** @internal Add RRSet copy to RR array. */
-int kr_rrarray_add(rr_array_t *array, const knot_rrset_t *rr, knot_mm_t *pool);
-
 /** @internal Add RRSet copy to ranked RR array. */
+KR_EXPORT
 int kr_ranked_rrarray_add(ranked_rr_array_t *array, const knot_rrset_t *rr,
 			  uint8_t rank, bool to_wire, uint32_t qry_uid, knot_mm_t *pool);
 
-int kr_ranked_rrarray_set_wire(ranked_rr_array_t *array, bool to_wire, uint32_t qry_uid);
+/** @internal Mark the RRSets from particular query as
+ * "have (not) to be recorded in the final answer".
+ * @param array RRSet array.
+ * @param to_wire Records must be\must not be recorded in final answer.
+ * @param qry_uid Query uid.
+ * @param check_dups When to_wire is true, try to avoid duplicate RRSets.
+ * @return 0 or an error
+ */
+int kr_ranked_rrarray_set_wire(ranked_rr_array_t *array, bool to_wire,
+			       uint32_t qry_uid, bool check_dups);
 
 void kr_rrset_print(const knot_rrset_t *rr, const char *prefix);
+void kr_qry_print(const struct kr_query *qry, const char *prefix, const char *postfix);
 void kr_pkt_print(knot_pkt_t *pkt);
 void kr_dname_print(const knot_dname_t *name, const char *prefix, const char *postfix);
 void kr_rrtype_print(const uint16_t rrtype, const char *prefix, const char *postfix);
@@ -211,3 +277,12 @@ char *kr_module_call(struct kr_context *ctx, const char *module, const char *pro
 	memcpy(&x, swap_temp, sizeof(x)); \
 	} while(0)
 
+/** Return the (covered) type of an nonempty RRset. */
+static inline uint16_t kr_rrset_type_maysig(const knot_rrset_t *rr)
+{
+	assert(rr && rr->rrs.rr_count && rr->rrs.data);
+	uint16_t type = rr->type;
+	if (type == KNOT_RRTYPE_RRSIG)
+		type = knot_rrsig_type_covered(&rr->rrs, 0);
+	return type;
+}
