@@ -343,15 +343,21 @@ static void fetch_addr(struct kr_zonecut *cut, struct kr_cache *cache, const kno
 }
 
 /** Fetch best NS for zone cut. */
-static int fetch_ns(struct kr_context *ctx, struct kr_zonecut *cut, const knot_dname_t *name, uint32_t timestamp, uint8_t * restrict rank)
+static int fetch_ns(struct kr_context *ctx, struct kr_zonecut *cut,
+		    const knot_dname_t *name, uint32_t timestamp,
+		    uint8_t * restrict rank, uint8_t * restrict flags)
 {
 	uint32_t drift = timestamp;
 	knot_rrset_t cached_rr;
 	knot_rrset_init(&cached_rr, (knot_dname_t *)name, KNOT_RRTYPE_NS, KNOT_CLASS_IN);
-	int ret = kr_cache_peek_rr(&ctx->cache, &cached_rr, rank, NULL, &drift);
+	int ret = kr_cache_peek_rr(&ctx->cache, &cached_rr, rank, flags, &drift);
 	if (ret != 0) {
 		return ret;
 	}
+	/* Note: we accept *any* rank from the cache.  We assume that nothing
+	 * completely untrustworthy could get into the cache, e.g out-of-bailiwick
+	 * records that weren't validated.
+	 */
 
 	/* Materialize as we'll going to do more cache lookups. */
 	knot_rrset_t rr_copy;
@@ -382,10 +388,10 @@ static int fetch_ns(struct kr_context *ctx, struct kr_zonecut *cut, const knot_d
 }
 
 /**
- * Fetch RRSet of given type.
+ * Fetch secure RRSet of given type.
  */
-static int fetch_rrset(knot_rrset_t **rr, struct kr_cache *cache,
-                       const knot_dname_t *owner, uint16_t type, knot_mm_t *pool, uint32_t timestamp)
+static int fetch_secure_rrset(knot_rrset_t **rr, struct kr_cache *cache,
+	const knot_dname_t *owner, uint16_t type, knot_mm_t *pool, uint32_t timestamp)
 {
 	if (!rr) {
 		return kr_error(ENOENT);
@@ -399,6 +405,10 @@ static int fetch_rrset(knot_rrset_t **rr, struct kr_cache *cache,
 	if (ret != 0) {
 		return ret;
 	}
+	const bool rankOK = kr_rank_test(rank, KR_RANK_SECURE);
+	if (!rankOK) {
+		return kr_error(ENOENT);
+	}
 
 	knot_rrset_free(rr, pool);
 	*rr = mm_alloc(pool, sizeof(knot_rrset_t));
@@ -415,21 +425,6 @@ static int fetch_rrset(knot_rrset_t **rr, struct kr_cache *cache,
 	return kr_ok();
 }
 
-/**
- * Fetch trust anchors for zone cut.
- * @note The trust anchor can theoretically be a DNSKEY but for now lets use only DS.
- */
-static int fetch_ta(struct kr_zonecut *cut, struct kr_cache *cache, const knot_dname_t *name, uint32_t timestamp)
-{
-	return fetch_rrset(&cut->trust_anchor, cache, name, KNOT_RRTYPE_DS, cut->pool, timestamp);
-}
-
-/** Fetch DNSKEY for zone cut. */
-static int fetch_dnskey(struct kr_zonecut *cut, struct kr_cache *cache, const knot_dname_t *name, uint32_t timestamp)
-{
-	return fetch_rrset(&cut->key, cache, name, KNOT_RRTYPE_DNSKEY, cut->pool, timestamp);
-}
-
 int kr_zonecut_find_cached(struct kr_context *ctx, struct kr_zonecut *cut, const knot_dname_t *name,
                            uint32_t timestamp, bool * restrict secured)
 {
@@ -446,15 +441,20 @@ int kr_zonecut_find_cached(struct kr_context *ctx, struct kr_zonecut *cut, const
 	while (true) {
 		/* Fetch NS first and see if it's insecure. */
 		uint8_t rank = 0;
+		uint8_t flags = 0;
 		const bool is_root = (label[0] == '\0');
-		if (fetch_ns(ctx, cut, label, timestamp, &rank) == 0) {
+		if (fetch_ns(ctx, cut, label, timestamp, &rank, &flags) == 0) {
 			/* Flag as insecure if cached as this */
-			if (rank & KR_RANK_INSECURE)
+			if (kr_rank_test(rank, KR_RANK_INSECURE) ||
+			    (flags & KR_CACHE_FLAG_NODS)) {
 				*secured = false;
-			/* Fetch DS if caller wants secure zone cut */
+			}
+			/* Fetch DS and DNSKEY if caller wants secure zone cut */
 			if (*secured || is_root) {
-				fetch_ta(cut, &ctx->cache, label, timestamp);
-				fetch_dnskey(cut, &ctx->cache, label, timestamp);
+				fetch_secure_rrset(&cut->trust_anchor, &ctx->cache, label,
+					    KNOT_RRTYPE_DS, cut->pool, timestamp);
+				fetch_secure_rrset(&cut->key, &ctx->cache, label,
+					    KNOT_RRTYPE_DNSKEY, cut->pool, timestamp);
 			}
 			update_cut_name(cut, label);
 			mm_free(cut->pool, qname);
