@@ -31,7 +31,7 @@ function predict.drain(ev)
 	local deleted = 0
 	for key, val in pairs(predict.queue) do
 		local qtype, qname = key:match('(%S*)%s(.*)')
-		worker.resolve(qname, kres.type[qtype], 1, kres.query.NO_CACHE)
+		worker.resolve(qname, kres.type[qtype], kres.class.IN, kres.query.NO_CACHE)
 		predict.queue[key] = nil
 		deleted = deleted + 1
 		if deleted >= predict.batch then
@@ -54,7 +54,7 @@ local function enqueue(queries)
 	local nr_queries = #queries
 	for i = 1, nr_queries do
 		local entry = queries[i]
-		local key = string.format('%s %s', entry.stype, entry.name)
+		local key = string.format('%s %s', entry.type, entry.name)
 		if not predict.queue[key] then
 			predict.queue[key] = 1
 			queued = queued + 1
@@ -63,44 +63,42 @@ local function enqueue(queries)
 	return queued	
 end
 
--- Prefetch soon-to-expire records
-function predict.prefetch()
-	local queries = stats.expiring()
-	stats.clear_expiring()
-	return enqueue(queries)
+-- Enqueue queries from same format as predict.queue or predict.log 
+local function enqueue_from_log(current)
+	if not current then return 0 end
+	queued = 0
+	for key, val in pairs(current) do 
+		if val and not predict.queue[key] then
+			predict.queue[key] = val
+			queued = queued + 1
+		end
+	end
+	return queued
 end
 
 -- Sample current epoch, return number of sampled queries
 function predict.sample(epoch_now)
 	if not epoch_now then return 0, 0 end
+	local current = predict.log[epoch_now] or {}	
 	local queries = stats.frequent()
 	stats.clear_frequent()
-	local queued = 0
-	local current = predict.log[epoch_now]
-	if predict.epoch ~= epoch_now or current == nil then
-		if current ~= nil then
-			queued = enqueue(current)
-		end
-		current = {}
-	end
 	local nr_samples = #queries
 	for i = 1, nr_samples do
 		local entry = queries[i]
-		local key = string.format('%s %s', entry.stype, entry.name)
+		local key = string.format('%s %s', entry.type, entry.name)
 		current[key] = 1
 	end
 	predict.log[epoch_now] = current
-	return nr_samples, queued
+	return nr_samples
 end
 
 -- Predict queries for the upcoming epoch
 local function generate(epoch_now)
 	if not epoch_now then return 0 end
 	local queued = 0
-	local period = predict.period + 1
 	for i = 1, predict.period / 2 - 1 do
-		local current = predict.log[(epoch_now - i) % period]
-		local past = predict.log[(epoch_now - 2*i) % period]
+		local current = predict.log[(epoch_now - i - 1) % predict.period + 1]
+		local past = predict.log[(epoch_now - 2*i - 1) % predict.period + 1]
 		if current and past then
 			for k, v in pairs(current) do
 				if past[k] ~= nil and not predict.queue[k] then
@@ -114,19 +112,29 @@ local function generate(epoch_now)
 end
 
 function predict.process(ev)
-	if not stats then error("'stats' module required") end
+	if (predict.period or 0) ~= 0 and not stats then
+		error("'stats' module required")
+	end
 	-- Start a new epoch, or continue sampling
 	predict.ev_sample = nil
 	local epoch_now = current_epoch()
-	local nr_learned, nr_queued = predict.sample(epoch_now)
-	-- End of epoch, predict next
+	local nr_queued = 0
+
+	-- End of epoch 
 	if predict.epoch ~= epoch_now then
 		stats['predict.epoch'] = epoch_now
 		predict.epoch = epoch_now
+		-- enqueue records from upcoming epoch	
+		nr_queued = enqueue_from_log(predict.log[epoch_now])
+		-- predict next epoch
 		nr_queued = nr_queued + generate(epoch_now)
+		-- clear log for new epoch
+		predict.log[epoch_now] = {}
 	end
-	-- Prefetch expiring records
-	nr_queued = nr_queued + predict.prefetch()
+	
+	-- Sample current epoch
+	local nr_learned = predict.sample(epoch_now)
+	
 	-- Dispatch predicted queries
 	if nr_queued > 0 then
 		predict.queue_len = predict.queue_len + nr_queued
@@ -136,8 +144,10 @@ function predict.process(ev)
 		end
 	end
 	predict.ev_sample = event.after(next_event(), predict.process)
-	stats['predict.queue'] = predict.queue_len
-	stats['predict.learned'] = nr_learned
+	if stats then
+		stats['predict.queue'] = predict.queue_len
+		stats['predict.learned'] = nr_learned
+	end
 	collectgarbage()
 end
 
@@ -169,4 +179,19 @@ function predict.config(config)
 	predict.init()
 end
 
+predict.layer = {
+	-- Prefetch all expiring (sub-)queries immediately after the request finishes.
+	-- Doing that immediately is simplest and avoids creating (new) large bursts of activity.
+	finish = function (state, req)
+		req = kres.request_t(req)
+		local qrys = req.rplan.resolved
+		for i = 0, (tonumber(qrys.len) - 1) do -- size_t doesn't work for some reason
+			local qry = qrys.at[i]
+			if bit.band(qry.flags, kres.query.EXPIRING) ~= 0 then
+				worker.resolve(kres.dname2str(qry.sname), qry.stype, qry.sclass, kres.query.NO_CACHE)
+			end
+		end
+	end
+}
+
 return predict
