@@ -7,7 +7,7 @@ CMOCKA_TAG="cmocka-1.1.0"
 CMOCKA_URL="git://git.cryptomilk.org/projects/cmocka.git"
 LIBUV_TAG="v1.10.2"
 LIBUV_URL="https://github.com/libuv/libuv.git"
-KNOT_TAG="v2.3.3"
+KNOT_TAG="v2.4.2"
 KNOT_URL="https://github.com/CZ-NIC/knot.git"
 GMP_TAG="6.1.1" # GMP 6.1.2 is broken on Travis-CI
 GMP_URL="https://gmplib.org/download/gmp/gmp-${GMP_TAG}.tar.xz"
@@ -23,6 +23,12 @@ HIREDIS_TAG="v0.13.3"
 HIREDIS_URL="https://github.com/redis/hiredis.git"
 LIBMEMCACHED_TAG="1.0.18"
 LIBMEMCACHED_URL="https://launchpad.net/libmemcached/1.0/${LIBMEMCACHED_TAG}/+download/libmemcached-${LIBMEMCACHED_TAG}.tar.gz"
+PROTOBUF_VER="2.6.1"
+PROTOBUF_URL="https://github.com/google/protobuf/archive/v${PROTOBUF_VER}.tar.gz"
+PROTOBUFC_VER="1.2.1"
+PROTOBUFC_URL="https://github.com/protobuf-c/protobuf-c/releases/download/v${PROTOBUFC_VER}/protobuf-c-${PROTOBUFC_VER}.tar.gz"
+FSTRM_VER="0.3.2"
+FSTRM_URL="https://github.com/farsightsec/fstrm/archive/v${FSTRM_VER}.tar.gz"
 
 if command -v shasum >/dev/null; then
     SHASUM="shasum -a 256"
@@ -38,11 +44,14 @@ PREFIX=${1}; [ -z ${PREFIX} ] && export PREFIX="${HOME}/.local"
 
 function bootstrap_cleanup {
     if [ -n "$BOOTSTRAP_CLEANUP" ]; then
-	echo "Bootstrap script has changed, cleaning up ${PREFIX}"
-	rm -rf "${PREFIX}"
+		echo "Bootstrap script has changed, cleaning up ${PREFIX}"
+		rm -rf "${PREFIX}"
     else
-	echo "Bootstrap script has changed, you should cleanup ${PREFIX}"
-	echo "or rerun this script with BOOSTRAP_CLEANUP=1 env variable"
+		echo "Bootstrap script has changed, you should cleanup ${PREFIX}"
+		echo "or rerun this script with BOOSTRAP_CLEANUP=1 env variable."
+		if [ "$PREFIX" = "$HOME/.local" ]; then
+			echo "BEWARE: e.g. your ~/.local/share may contain something unrelated."
+		fi
     fi
 }
 
@@ -121,9 +130,11 @@ function pkg {
 PIP_PKGS="dnspython==1.11 cpp-coveralls Jinja2"
 if [ "${TRAVIS_OS_NAME}" == "osx" ]; then
 	brew update
-	brew install --force makedepend python hiredis libmemcached || true
-	brew link --overwrite python || true
-	pip install --upgrade pip || true
+	for p in makedepend hiredis libmemcached protobuf-c cmocka jansson gnutls luajit libuv; do
+		echo "BEGIN $p";
+		brew install $p
+		echo "END $p";
+	done
 	pip install ${PIP_PKGS}
 fi
 if [ "${TRAVIS_OS_NAME}" == "linux" ]; then
@@ -132,22 +143,25 @@ if [ "${TRAVIS_OS_NAME}" == "linux" ]; then
 	rm ${HOME}/.cache/pip/log/debug.log || true
 	pkg hiredis ${HIREDIS_URL} ${HIREDIS_TAG} include/hiredis/hiredis.h install PREFIX=${PREFIX}
 	pkg libmemcached ${LIBMEMCACHED_URL} ${LIBMEMCACHED_TAG} include/libmemcached/memcached.h
+	pkg cmocka ${CMOCKA_URL} ${CMOCKA_TAG} include/cmocka.h
+	pkg protobuf ${PROTOBUF_URL} ${PROTOBUF_VER} lib/pkgconfig/protobuf.pc
+	pkg protobuf-c ${PROTOBUFC_URL} ${PROTOBUFC_VER} include/protobuf-c/protobuf-c.h
+	pkg jansson ${JANSSON_URL} ${JANSSON_TAG} include/jansson.h --disable-static
+	pkg gmp ${GMP_URL} ${GMP_TAG} include/gmp.h --disable-static
+	pkg nettle ${NETTLE_URL} ${NETTLE_TAG} include/nettle \
+		--disable-documentation --with-lib-path=${PREFIX}/lib --with-include-path=${PREFIX}/include
+	export GMP_CFLAGS="-I${PREFIX}/include"
+	export GMP_LIBS="-L${PREFIX}/lib -lgmp"
+	pkg gnutls ${GNUTLS_URL} ${GNUTLS_TAG} include/gnutls \
+	    --disable-tests --disable-doc --disable-valgrind-tests --disable-static --with-included-libtasn1 --without-p11-kit \
+	    --disable-tools --disable-cxx --with-included-unistring
+	pkg lua ${LUA_URL} v${LUA_VER} lib/pkgconfig/luajit.pc install BUILDMODE=dynamic LDFLAGS=-lm PREFIX=${PREFIX}
+	pkg libuv ${LIBUV_URL} ${LIBUV_TAG} include/uv.h --disable-static
 fi
 
-pkg gmp ${GMP_URL} ${GMP_TAG} include/gmp.h --disable-static
-pkg nettle ${NETTLE_URL} ${NETTLE_TAG} include/nettle \
-	--disable-documentation --with-lib-path=${PREFIX}/lib --with-include-path=${PREFIX}/include
-export GMP_CFLAGS="-I${PREFIX}/include"
-export GMP_LIBS="-L${PREFIX}/lib -lgmp"
-pkg gnutls ${GNUTLS_URL} ${GNUTLS_TAG} include/gnutls \
-    --disable-tests --disable-doc --disable-valgrind-tests --disable-static --with-included-libtasn1 --without-p11-kit \
-    --disable-tools --disable-cxx --with-included-unistring
-pkg jansson ${JANSSON_URL} ${JANSSON_TAG} include/jansson.h --disable-static
 pkg libknot ${KNOT_URL} ${KNOT_TAG} include/libknot \
 	--disable-static --with-lmdb=no --disable-fastparser --disable-daemon --disable-utilities --disable-documentation
-pkg cmocka ${CMOCKA_URL} ${CMOCKA_TAG} include/cmocka.h
-pkg libuv ${LIBUV_URL} ${LIBUV_TAG} include/uv.h --disable-static
-pkg lua ${LUA_URL} v${LUA_VER} lib/pkgconfig/luajit.pc install BUILDMODE=dynamic LDFLAGS=-lm PREFIX=${PREFIX}
+pkg fstrm ${FSTRM_URL} ${FSTRM_VER} include/fstrm.h --disable-programs
 
 # development releases of luajit do NOT install bin/luajit
 ln -sf "luajit-${LUA_VER}" "${PREFIX}/bin/luajit"
